# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes

import ddt
import mock

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.compute import _clusapi_utils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi.libs import clusapi as clusapi_def
from os_win.utils.winapi import wintypes


@ddt.ddt
class ClusApiUtilsTestCase(test_base.OsWinBaseTestCase):
    _LIVE_MIGRATION_TYPE = 4

    def setUp(self):
        super(ClusApiUtilsTestCase, self).setUp()

        self._clusapi = mock.patch.object(
            _clusapi_utils, 'clusapi', create=True).start()

        self._clusapi_utils = _clusapi_utils.ClusApiUtils()

        self._run_patcher = mock.patch.object(self._clusapi_utils,
                                              '_run_and_check_output')
        self._mock_run = self._run_patcher.start()

    def _mock_ctypes(self):
        self._ctypes = mock.Mock()
        # This is used in order to easily make assertions on the variables
        # passed by reference.
        self._ctypes.byref = lambda x: (x, "byref")
        self._ctypes.c_wchar_p = lambda x: (x, 'c_wchar_p')
        self._ctypes.sizeof = lambda x: (x, 'sizeof')
        self._ctypes.c_ulong = lambda x: (x, 'c_ulong')

        mock.patch.object(_clusapi_utils, 'ctypes', self._ctypes).start()

    def test_run_and_check_output(self):
        self._clusapi_utils._win32utils = mock.Mock()
        self._clusapi_utils._run_and_check_output = (
            self._run_patcher.temp_original)

        mock_win32utils_run_and_check_output = (
            self._clusapi_utils._win32utils.run_and_check_output)

        ret_val = self._clusapi_utils._run_and_check_output(
            mock.sentinel.func,
            mock.sentinel.arg,
            fake_kwarg=mock.sentinel.kwarg)

        mock_win32utils_run_and_check_output.assert_called_once_with(
            mock.sentinel.func,
            mock.sentinel.arg,
            fake_kwarg=mock.sentinel.kwarg,
            failure_exc=exceptions.ClusterWin32Exception)
        self.assertEqual(mock_win32utils_run_and_check_output.return_value,
                         ret_val)

    def test_dword_align(self):
        self.assertEqual(8, self._clusapi_utils._dword_align(5))
        self.assertEqual(4, self._clusapi_utils._dword_align(4))

    def test_get_clusprop_value_struct(self):
        val_type = ctypes.c_ubyte * 3
        expected_padding_sz = 1

        clusprop_val_struct = self._clusapi_utils._get_clusprop_value_struct(
            val_type)

        expected_fields = [('syntax', wintypes.DWORD),
                           ('length', wintypes.DWORD),
                           ('value', val_type),
                           ('_padding', ctypes.c_ubyte * expected_padding_sz)]
        self.assertEqual(expected_fields, clusprop_val_struct._fields_)

    def test_get_property_list_entry(self):
        fake_prop_name = 'fake prop name'
        fake_prop_syntax = 1
        fake_prop_val = (ctypes.c_wchar * 10)()
        fake_prop_val.value = 'fake prop'

        entry = self._clusapi_utils.get_property_list_entry(
            name=fake_prop_name,
            syntax=fake_prop_syntax,
            value=fake_prop_val)

        self.assertEqual(w_const.CLUSPROP_SYNTAX_NAME,
                         entry.name.syntax)
        self.assertEqual(fake_prop_name,
                         entry.name.value)
        self.assertEqual(
            ctypes.sizeof(ctypes.c_wchar) * (len(fake_prop_name) + 1),
            entry.name.length)

        self.assertEqual(fake_prop_syntax,
                         entry.value.syntax)
        self.assertEqual(bytearray(fake_prop_val),
                         bytearray(entry.value.value))
        self.assertEqual(
            ctypes.sizeof(fake_prop_val),
            entry.value.length)

        self.assertEqual(w_const.CLUSPROP_SYNTAX_ENDMARK,
                         entry._endmark)

    def test_get_property_list(self):
        entry_0 = self._clusapi_utils.get_property_list_entry(
            name='fake prop name',
            syntax=1,
            value=ctypes.c_uint(2))
        entry_1 = self._clusapi_utils.get_property_list_entry(
            name='fake prop name',
            syntax=2,
            value=ctypes.c_ubyte(5))

        prop_list = self._clusapi_utils.get_property_list(
            [entry_0, entry_1])

        self.assertEqual(2, prop_list.count)
        self.assertEqual(bytearray(entry_0) + bytearray(entry_1),
                         prop_list.entries_buff)

    @ddt.data('fake cluster name', None)
    def test_open_cluster(self, cluster_name):
        self._mock_ctypes()

        handle = self._clusapi_utils.open_cluster(cluster_name)

        expected_handle_arg = (
            self._ctypes.c_wchar_p(cluster_name)
            if cluster_name else None)
        self._mock_run.assert_called_once_with(
            self._clusapi.OpenCluster,
            expected_handle_arg,
            **self._clusapi_utils._open_handle_check_flags)

        self.assertEqual(self._mock_run.return_value, handle)

    def test_open_cluster_enum(self):
        handle = self._clusapi_utils.open_cluster_enum(
            mock.sentinel.cluster_handle,
            mock.sentinel.object_type)

        self._mock_run.assert_called_once_with(
            self._clusapi.ClusterOpenEnumEx,
            mock.sentinel.cluster_handle,
            mock.sentinel.object_type,
            None,
            **self._clusapi_utils._open_handle_check_flags)

        self.assertEqual(self._mock_run.return_value, handle)

    def test_open_cluster_group(self):
        self._mock_ctypes()

        handle = self._clusapi_utils.open_cluster_group(
            mock.sentinel.cluster_handle,
            mock.sentinel.group_name)

        self._mock_run.assert_called_once_with(
            self._clusapi.OpenClusterGroup,
            mock.sentinel.cluster_handle,
            self._ctypes.c_wchar_p(mock.sentinel.group_name),
            **self._clusapi_utils._open_handle_check_flags)

        self.assertEqual(self._mock_run.return_value, handle)

    def test_open_cluster_node(self):
        self._mock_ctypes()

        handle = self._clusapi_utils.open_cluster_node(
            mock.sentinel.cluster_handle,
            mock.sentinel.node_name)

        self._mock_run.assert_called_once_with(
            self._clusapi.OpenClusterNode,
            mock.sentinel.cluster_handle,
            self._ctypes.c_wchar_p(mock.sentinel.node_name),
            **self._clusapi_utils._open_handle_check_flags)

        self.assertEqual(self._mock_run.return_value, handle)

    def test_open_cluster_resource(self):
        self._mock_ctypes()

        handle = self._clusapi_utils.open_cluster_resource(
            mock.sentinel.cluster_handle,
            mock.sentinel.resource_name)

        self._mock_run.assert_called_once_with(
            self._clusapi.OpenClusterResource,
            mock.sentinel.cluster_handle,
            self._ctypes.c_wchar_p(mock.sentinel.resource_name),
            **self._clusapi_utils._open_handle_check_flags)

        self.assertEqual(self._mock_run.return_value, handle)

    def test_close_cluster(self):
        self._clusapi_utils.close_cluster(mock.sentinel.handle)
        self._clusapi.CloseCluster.assert_called_once_with(
            mock.sentinel.handle)

    def test_close_cluster_group(self):
        self._clusapi_utils.close_cluster_group(mock.sentinel.handle)
        self._clusapi.CloseClusterGroup.assert_called_once_with(
            mock.sentinel.handle)

    def test_close_cluster_node(self):
        self._clusapi_utils.close_cluster_node(mock.sentinel.handle)
        self._clusapi.CloseClusterNode.assert_called_once_with(
            mock.sentinel.handle)

    def test_close_cluster_resource(self):
        self._clusapi_utils.close_cluster_resource(mock.sentinel.handle)
        self._clusapi.CloseClusterResource.assert_called_once_with(
            mock.sentinel.handle)

    def test_close_cluster_enum(self):
        self._clusapi_utils.close_cluster_enum(mock.sentinel.handle)
        self._clusapi.ClusterCloseEnumEx.assert_called_once_with(
            mock.sentinel.handle)

    def test_online_cluster_group(self):
        self._clusapi_utils.online_cluster_group(mock.sentinel.group_handle,
                                                 mock.sentinel.dest_handle)
        self._mock_run.assert_called_once_with(
            self._clusapi.OnlineClusterGroup,
            mock.sentinel.group_handle,
            mock.sentinel.dest_handle)

    def test_destroy_cluster_group(self):
        self._clusapi_utils.destroy_cluster_group(mock.sentinel.group_handle)
        self._mock_run.assert_called_once_with(
            self._clusapi.DestroyClusterGroup,
            mock.sentinel.group_handle)

    def test_offline_cluster_group(self):
        self._clusapi_utils.offline_cluster_group(mock.sentinel.group_handle)
        self._mock_run.assert_called_once_with(
            self._clusapi.OfflineClusterGroup,
            mock.sentinel.group_handle)

    @ddt.data(0, w_const.ERROR_IO_PENDING)
    def test_cancel_cluster_group_operation(self, cancel_ret_val):
        self._mock_run.return_value = cancel_ret_val

        expected_ret_val = cancel_ret_val != w_const.ERROR_IO_PENDING
        ret_val = self._clusapi_utils.cancel_cluster_group_operation(
            mock.sentinel.group_handle)

        self.assertEqual(expected_ret_val, ret_val)

        self._mock_run.assert_called_once_with(
            self._clusapi.CancelClusterGroupOperation,
            mock.sentinel.group_handle,
            0,
            ignored_error_codes=[w_const.ERROR_IO_PENDING])

    @ddt.data(mock.sentinel.prop_list, None)
    def test_move_cluster_group(self, prop_list):
        self._mock_ctypes()

        expected_prop_list_arg = (
            self._ctypes.byref(prop_list) if prop_list else None)
        expected_prop_list_sz = (
            self._ctypes.sizeof(prop_list) if prop_list else 0)

        self._clusapi_utils.move_cluster_group(
            mock.sentinel.group_handle,
            mock.sentinel.dest_node_handle,
            mock.sentinel.move_flags,
            prop_list)

        self._mock_run.assert_called_once_with(
            self._clusapi.MoveClusterGroupEx,
            mock.sentinel.group_handle,
            mock.sentinel.dest_node_handle,
            mock.sentinel.move_flags,
            expected_prop_list_arg,
            expected_prop_list_sz,
            ignored_error_codes=[w_const.ERROR_IO_PENDING])

    def test_get_cluster_group_state(self):
        owner_node = 'fake owner node'

        def fake_get_state(inst,
                           group_handle, node_name_buff, node_name_len,
                           error_ret_vals, error_on_nonzero_ret_val,
                           ret_val_is_err_code):
            self.assertEqual(mock.sentinel.group_handle, group_handle)

            # Those arguments would not normally get to the ClusApi
            # function, instead being used by the helper invoking
            # it and catching errors. For convenience, we validate
            # those arguments at this point.
            self.assertEqual([constants.CLUSTER_GROUP_STATE_UNKNOWN],
                             error_ret_vals)
            self.assertFalse(error_on_nonzero_ret_val)
            self.assertFalse(ret_val_is_err_code)

            node_name_len_arg = ctypes.cast(
                node_name_len,
                wintypes.PDWORD).contents
            self.assertEqual(w_const.MAX_PATH,
                             node_name_len_arg.value)

            node_name_arg = ctypes.cast(
                node_name_buff,
                ctypes.POINTER(
                    ctypes.c_wchar *
                    w_const.MAX_PATH)).contents
            node_name_arg.value = owner_node
            return mock.sentinel.group_state

        self._mock_run.side_effect = fake_get_state

        state_info = self._clusapi_utils.get_cluster_group_state(
            mock.sentinel.group_handle)
        expected_state_info = dict(state=mock.sentinel.group_state,
                                   owner_node=owner_node)
        self.assertEqual(expected_state_info, state_info)

    @ddt.data({'notif_filters': (clusapi_def.NOTIFY_FILTER_AND_TYPE * 2)(),
               'exp_notif_filters_len': 2},
              {'notif_filters': clusapi_def.NOTIFY_FILTER_AND_TYPE(),
               'notif_port_h': mock.sentinel.notif_port_h,
               'notif_key': mock.sentinel.notif_key})
    @ddt.unpack
    def test_create_cluster_notify_port(self, notif_filters,
                                        exp_notif_filters_len=1,
                                        notif_port_h=None,
                                        notif_key=None):
        self._mock_ctypes()
        self._ctypes.Array = ctypes.Array

        self._clusapi_utils.create_cluster_notify_port_v2(
            mock.sentinel.cluster_handle,
            notif_filters,
            notif_port_h,
            notif_key)

        exp_notif_key_p = self._ctypes.byref(notif_key) if notif_key else None
        exp_notif_port_h = notif_port_h or w_const.INVALID_HANDLE_VALUE

        self._mock_run.assert_called_once_with(
            self._clusapi.CreateClusterNotifyPortV2,
            exp_notif_port_h,
            mock.sentinel.cluster_handle,
            self._ctypes.byref(notif_filters),
            self._ctypes.c_ulong(exp_notif_filters_len),
            exp_notif_key_p,
            **self._clusapi_utils._open_handle_check_flags)

    def test_close_cluster_notify_port(self):
        self._clusapi_utils.close_cluster_notify_port(mock.sentinel.handle)
        self._clusapi.CloseClusterNotifyPort.assert_called_once_with(
            mock.sentinel.handle)

    def test_get_cluster_notify_v2(self):
        fake_notif_key = 1
        fake_notif_port_h = 2
        fake_notif_type = 3
        fake_filter_flags = 4
        fake_clus_obj_name = 'fake-changed-clus-object'
        fake_event_buff = 'fake-event-buff'
        fake_obj_type = 'fake-object-type'
        fake_obj_id = 'fake-obj-id'
        fake_parent_id = 'fake-parent-id'

        notif_key = ctypes.c_ulong(fake_notif_key)
        requested_buff_sz = 1024

        def fake_get_cluster_notify(func, notif_port_h, pp_notif_key,
                                    p_filter_and_type,
                                    p_buff, p_buff_sz,
                                    p_obj_id_buff, p_obj_id_buff_sz,
                                    p_parent_id_buff, p_parent_id_buff_sz,
                                    p_obj_name_buff, p_obj_name_buff_sz,
                                    p_obj_type, p_obj_type_sz,
                                    timeout_ms):
            self.assertEqual(self._clusapi.GetClusterNotifyV2, func)
            self.assertEqual(fake_notif_port_h, notif_port_h)

            obj_name_buff_sz = ctypes.cast(
                p_obj_name_buff_sz,
                wintypes.PDWORD).contents
            buff_sz = ctypes.cast(
                p_buff_sz,
                wintypes.PDWORD).contents
            obj_type_sz = ctypes.cast(
                p_obj_type_sz,
                wintypes.PDWORD).contents
            obj_id_sz = ctypes.cast(
                p_obj_id_buff_sz,
                wintypes.PDWORD).contents
            parent_id_buff_sz = ctypes.cast(
                p_parent_id_buff_sz,
                wintypes.PDWORD).contents

            # We'll just request the tested method to pass us
            # a buffer this large.
            if (buff_sz.value < requested_buff_sz or
                    obj_name_buff_sz.value < requested_buff_sz or
                    parent_id_buff_sz.value < requested_buff_sz or
                    obj_type_sz.value < requested_buff_sz or
                    obj_id_sz.value < requested_buff_sz):
                buff_sz.value = requested_buff_sz
                obj_name_buff_sz.value = requested_buff_sz
                parent_id_buff_sz.value = requested_buff_sz
                obj_type_sz.value = requested_buff_sz
                obj_id_sz.value = requested_buff_sz
                raise exceptions.ClusterWin32Exception(
                    error_code=w_const.ERROR_MORE_DATA,
                    func_name='GetClusterNotifyV2',
                    error_message='error more data')

            pp_notif_key = ctypes.cast(pp_notif_key, ctypes.c_void_p)
            p_notif_key = ctypes.c_void_p.from_address(pp_notif_key.value)
            p_notif_key.value = ctypes.addressof(notif_key)

            filter_and_type = ctypes.cast(
                p_filter_and_type,
                ctypes.POINTER(clusapi_def.NOTIFY_FILTER_AND_TYPE)).contents
            filter_and_type.dwObjectType = fake_notif_type
            filter_and_type.FilterFlags = fake_filter_flags

            def set_wchar_buff(p_wchar_buff, wchar_buff_sz, value):
                wchar_buff = ctypes.cast(
                    p_wchar_buff,
                    ctypes.POINTER(
                        ctypes.c_wchar *
                        (wchar_buff_sz // ctypes.sizeof(ctypes.c_wchar))))
                wchar_buff = wchar_buff.contents
                ctypes.memset(wchar_buff, 0, wchar_buff_sz)
                wchar_buff.value = value
                return wchar_buff

            set_wchar_buff(p_obj_name_buff, requested_buff_sz,
                           fake_clus_obj_name)
            set_wchar_buff(p_buff, requested_buff_sz, fake_event_buff)
            set_wchar_buff(p_parent_id_buff, requested_buff_sz, fake_parent_id)
            set_wchar_buff(p_obj_type, requested_buff_sz, fake_obj_type)
            set_wchar_buff(p_obj_id_buff, requested_buff_sz, fake_obj_id)

            self.assertEqual(mock.sentinel.timeout_ms, timeout_ms)

        self._mock_run.side_effect = fake_get_cluster_notify

        event = self._clusapi_utils.get_cluster_notify_v2(
            fake_notif_port_h, mock.sentinel.timeout_ms)
        w_event_buff = ctypes.cast(
            event['buff'],
            ctypes.POINTER(
                ctypes.c_wchar *
                (requested_buff_sz // ctypes.sizeof(ctypes.c_wchar))))
        w_event_buff = w_event_buff.contents[:]
        event['buff'] = w_event_buff.split('\x00')[0]

        expected_event = dict(cluster_object_name=fake_clus_obj_name,
                              object_id=fake_obj_id,
                              object_type=fake_notif_type,
                              object_type_str=fake_obj_type,
                              filter_flags=fake_filter_flags,
                              parent_id=fake_parent_id,
                              buff=fake_event_buff,
                              buff_sz=requested_buff_sz,
                              notif_key=fake_notif_key)
        self.assertEqual(expected_event, event)

    def _get_fake_prop_list(self):
        syntax = w_const.CLUSPROP_SYNTAX_LIST_VALUE_DWORD
        migr_type = wintypes.DWORD(self._LIVE_MIGRATION_TYPE)

        prop_entries = [
            self._clusapi_utils.get_property_list_entry(
                w_const.CLUS_RESTYPE_NAME_VM, syntax, migr_type),
            self._clusapi_utils.get_property_list_entry(
                w_const.CLUS_RESTYPE_NAME_VM_CONFIG, syntax, migr_type),
            self._clusapi_utils.get_property_list_entry(
                w_const.CLUSREG_NAME_GRP_STATUS_INFORMATION,
                w_const.CLUSPROP_SYNTAX_LIST_VALUE_ULARGE_INTEGER,
                ctypes.c_ulonglong(w_const.
                    CLUSGRP_STATUS_WAITING_IN_QUEUE_FOR_MOVE)),  # noqa
            self._clusapi_utils.get_property_list_entry(
                w_const.CLUSREG_NAME_GRP_TYPE,
                w_const.CLUSPROP_SYNTAX_LIST_VALUE_DWORD,
                ctypes.c_ulonglong(w_const.ClusGroupTypeVirtualMachine)),
        ]

        prop_list = self._clusapi_utils.get_property_list(prop_entries)
        return prop_list

    def test_get_prop_list_entry_p_not_found(self):
        prop_list = self._get_fake_prop_list()

        self.assertRaises(exceptions.ClusterPropertyListEntryNotFound,
                          self._clusapi_utils.get_prop_list_entry_p,
                          ctypes.byref(prop_list),
                          ctypes.sizeof(prop_list),
                          'InexistentProperty')

    def test_get_prop_list_entry_p_parsing_error(self):
        prop_list = self._get_fake_prop_list()

        prop_entry_name_len_addr = ctypes.addressof(
            prop_list.entries_buff) + ctypes.sizeof(ctypes.c_ulong)
        prop_entry_name_len = ctypes.c_ulong.from_address(
            prop_entry_name_len_addr)
        prop_entry_name_len.value = ctypes.sizeof(prop_list)

        self.assertRaises(exceptions.ClusterPropertyListParsingError,
                          self._clusapi_utils.get_prop_list_entry_p,
                          ctypes.byref(prop_list),
                          ctypes.sizeof(prop_list),
                          w_const.CLUS_RESTYPE_NAME_VM)

    def test_get_prop_list_entry_p(self):
        prop_list = self._get_fake_prop_list()

        prop_entry = self._clusapi_utils.get_prop_list_entry_p(
            ctypes.byref(prop_list),
            ctypes.sizeof(prop_list),
            w_const.CLUS_RESTYPE_NAME_VM_CONFIG)

        self.assertEqual(
            w_const.CLUSPROP_SYNTAX_LIST_VALUE_DWORD,
            prop_entry['syntax'])
        self.assertEqual(
            ctypes.sizeof(ctypes.c_ulong),
            prop_entry['length'])

        val = ctypes.c_ulong.from_address(prop_entry['val_p'].value).value
        self.assertEqual(self._LIVE_MIGRATION_TYPE, val)

    def test_cluster_group_control(self):
        fake_out_buff = 'fake-event-buff'

        requested_buff_sz = 1024

        def fake_cluster_group_ctrl(func, group_handle, node_handle,
                                    control_code,
                                    in_buff_p, in_buff_sz,
                                    out_buff_p, out_buff_sz,
                                    requested_buff_sz_p):
            self.assertEqual(self._clusapi.ClusterGroupControl, func)
            self.assertEqual(mock.sentinel.group_handle, group_handle)
            self.assertEqual(mock.sentinel.node_handle, node_handle)
            self.assertEqual(mock.sentinel.control_code, control_code)
            self.assertEqual(mock.sentinel.in_buff_p, in_buff_p)
            self.assertEqual(mock.sentinel.in_buff_sz, in_buff_sz)

            req_buff_sz = ctypes.cast(
                requested_buff_sz_p,
                wintypes.PDWORD).contents
            req_buff_sz.value = requested_buff_sz

            # We'll just request the tested method to pass us
            # a buffer this large.
            if (out_buff_sz.value < requested_buff_sz):
                raise exceptions.ClusterWin32Exception(
                    error_code=w_const.ERROR_MORE_DATA,
                    func_name='ClusterGroupControl',
                    error_message='error more data')

            out_buff = ctypes.cast(
                out_buff_p,
                ctypes.POINTER(
                    ctypes.c_wchar *
                    (requested_buff_sz // ctypes.sizeof(ctypes.c_wchar))))
            out_buff = out_buff.contents
            out_buff.value = fake_out_buff

        self._mock_run.side_effect = fake_cluster_group_ctrl

        out_buff, out_buff_sz = self._clusapi_utils.cluster_group_control(
            mock.sentinel.group_handle, mock.sentinel.control_code,
            mock.sentinel.node_handle, mock.sentinel.in_buff_p,
            mock.sentinel.in_buff_sz)

        self.assertEqual(requested_buff_sz, out_buff_sz)
        wp_out_buff = ctypes.cast(
            out_buff,
            ctypes.POINTER(ctypes.c_wchar * requested_buff_sz))
        self.assertEqual(fake_out_buff,
                         wp_out_buff.contents[:len(fake_out_buff)])

    def test_get_cluster_group_status_info(self):
        prop_list = self._get_fake_prop_list()

        status_info = self._clusapi_utils.get_cluster_group_status_info(
            ctypes.byref(prop_list), ctypes.sizeof(prop_list))
        self.assertEqual(
            w_const.CLUSGRP_STATUS_WAITING_IN_QUEUE_FOR_MOVE,
            status_info)

    def test_get_cluster_group_type(self):
        prop_list = self._get_fake_prop_list()

        status_info = self._clusapi_utils.get_cluster_group_type(
            ctypes.byref(prop_list), ctypes.sizeof(prop_list))
        self.assertEqual(
            w_const.ClusGroupTypeVirtualMachine,
            status_info)

    def test_cluster_get_enum_count(self):
        ret_val = self._clusapi_utils.cluster_get_enum_count(
            mock.sentinel.enum_handle)

        self.assertEqual(self._mock_run.return_value, ret_val)
        self._mock_run.assert_called_once_with(
            self._clusapi.ClusterGetEnumCountEx,
            mock.sentinel.enum_handle,
            error_on_nonzero_ret_val=False,
            ret_val_is_err_code=False)

    def test_cluster_enum(self):
        obj_id = 'fake_obj_id'
        obj_id_wchar_p = ctypes.c_wchar_p(obj_id)

        requested_buff_sz = 1024

        def fake_cluster_enum(func, enum_handle, index, buff_p, buff_sz_p,
                              ignored_error_codes=tuple()):
            self.assertEqual(self._clusapi.ClusterEnumEx, func)
            self.assertEqual(mock.sentinel.enum_handle, enum_handle)
            self.assertEqual(mock.sentinel.index, index)

            buff_sz = ctypes.cast(
                buff_sz_p,
                wintypes.PDWORD).contents
            # We'll just request the tested method to pass us
            # a buffer this large.
            if (buff_sz.value < requested_buff_sz):
                buff_sz.value = requested_buff_sz
                if w_const.ERROR_MORE_DATA not in ignored_error_codes:
                    raise exceptions.ClusterWin32Exception(
                        error_code=w_const.ERROR_MORE_DATA)
                return

            item = ctypes.cast(
                buff_p,
                clusapi_def.PCLUSTER_ENUM_ITEM).contents
            item.lpszId = obj_id_wchar_p
            item.cbId = len(obj_id)

        self._mock_run.side_effect = fake_cluster_enum

        item = self._clusapi_utils.cluster_enum(
            mock.sentinel.enum_handle, mock.sentinel.index)
        self.assertEqual(obj_id, item.lpszId)


@ddt.ddt
class TestClusterContextManager(test_base.OsWinBaseTestCase):
    _autospec_classes = [_clusapi_utils.ClusApiUtils]

    def setUp(self):
        super(TestClusterContextManager, self).setUp()

        self._cmgr = _clusapi_utils.ClusterContextManager()
        self._clusapi_utils = self._cmgr._clusapi_utils

    @ddt.data(None, mock.sentinel.cluster_name)
    def test_open_cluster(self, cluster_name):
        with self._cmgr.open_cluster(cluster_name) as f:
            self._clusapi_utils.open_cluster.assert_called_once_with(
                cluster_name)
            self.assertEqual(f, self._clusapi_utils.open_cluster.return_value)

        self._clusapi_utils.close_cluster.assert_called_once_with(
            self._clusapi_utils.open_cluster.return_value)

    def test_open_cluster_group(self):
        with self._cmgr.open_cluster_group(mock.sentinel.group_name) as f:
            self._clusapi_utils.open_cluster.assert_called_once_with(None)
            self._clusapi_utils.open_cluster_group.assert_called_once_with(
                self._clusapi_utils.open_cluster.return_value,
                mock.sentinel.group_name)

            self.assertEqual(
                f,
                self._clusapi_utils.open_cluster_group.return_value)

        self._clusapi_utils.close_cluster_group.assert_called_once_with(
            self._clusapi_utils.open_cluster_group.return_value)
        self._clusapi_utils.close_cluster.assert_called_once_with(
            self._clusapi_utils.open_cluster.return_value)

    def test_open_missing_cluster_group(self):
        exc = exceptions.ClusterWin32Exception(
            func_name='OpenClusterGroup',
            message='expected exception',
            error_code=w_const.ERROR_GROUP_NOT_FOUND)
        self._clusapi_utils.open_cluster_group.side_effect = exc

        self.assertRaises(
            exceptions.ClusterObjectNotFound,
            self._cmgr.open_cluster_group(mock.sentinel.group_name).__enter__)

    def test_open_cluster_group_with_handle(self):
        with self._cmgr.open_cluster_group(
                mock.sentinel.group_name,
                cluster_handle=mock.sentinel.cluster_handle) as f:
            self._clusapi_utils.open_cluster.assert_not_called()
            self._clusapi_utils.open_cluster_group.assert_called_once_with(
                mock.sentinel.cluster_handle, mock.sentinel.group_name)

            self.assertEqual(
                f,
                self._clusapi_utils.open_cluster_group.return_value)

        self._clusapi_utils.close_cluster_group.assert_called_once_with(
            self._clusapi_utils.open_cluster_group.return_value)
        # If we pass our own handle, we don't want the tested method to
        # close it.
        self._clusapi_utils.close_cluster.assert_not_called()

    def test_open_cluster_resource(self):
        with self._cmgr.open_cluster_resource(mock.sentinel.res_name) as f:
            self._clusapi_utils.open_cluster.assert_called_once_with(None)
            self._clusapi_utils.open_cluster_resource.assert_called_once_with(
                self._clusapi_utils.open_cluster.return_value,
                mock.sentinel.res_name)

            self.assertEqual(
                f,
                self._clusapi_utils.open_cluster_resource.return_value)

        self._clusapi_utils.close_cluster_resource.assert_called_once_with(
            self._clusapi_utils.open_cluster_resource.return_value)
        self._clusapi_utils.close_cluster.assert_called_once_with(
            self._clusapi_utils.open_cluster.return_value)

    def test_open_cluster_node(self):
        with self._cmgr.open_cluster_node(mock.sentinel.node_name) as f:
            self._clusapi_utils.open_cluster.assert_called_once_with(None)
            self._clusapi_utils.open_cluster_node.assert_called_once_with(
                self._clusapi_utils.open_cluster.return_value,
                mock.sentinel.node_name)

            self.assertEqual(
                f,
                self._clusapi_utils.open_cluster_node.return_value)

        self._clusapi_utils.close_cluster_node.assert_called_once_with(
            self._clusapi_utils.open_cluster_node.return_value)
        self._clusapi_utils.close_cluster.assert_called_once_with(
            self._clusapi_utils.open_cluster.return_value)

    def test_open_cluster_enum(self):
        with self._cmgr.open_cluster_enum(mock.sentinel.object_type) as f:
            self._clusapi_utils.open_cluster.assert_called_once_with(None)
            self._clusapi_utils.open_cluster_enum.assert_called_once_with(
                self._clusapi_utils.open_cluster.return_value,
                mock.sentinel.object_type)

            self.assertEqual(
                f,
                self._clusapi_utils.open_cluster_enum.return_value)

        self._clusapi_utils.close_cluster_enum.assert_called_once_with(
            self._clusapi_utils.open_cluster_enum.return_value)
        self._clusapi_utils.close_cluster.assert_called_once_with(
            self._clusapi_utils.open_cluster.return_value)

    def test_invalid_handle_type(self):
        self.assertRaises(exceptions.Invalid,
                          self._cmgr._open(handle_type=None).__enter__)
        self.assertRaises(exceptions.Invalid,
                          self._cmgr._close, mock.sentinel.handle,
                          handle_type=None)



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import socket

from oslo_log import log as logging

from os_win._i18n import _
from os_win import _utils
from os_win import exceptions
from os_win.utils import baseutils
from os_win.utils import win32utils

LOG = logging.getLogger(__name__)


class SMBUtils(baseutils.BaseUtils):
    _loopback_share_map = {}

    def __init__(self):
        self._win32_utils = win32utils.Win32Utils()
        self._smb_conn = self._get_wmi_conn(r"root\Microsoft\Windows\SMB")

    def check_smb_mapping(self, share_path, remove_unavailable_mapping=False):
        mappings = self._smb_conn.Msft_SmbMapping(RemotePath=share_path)

        if not mappings:
            return False

        if os.path.exists(share_path):
            LOG.debug('Share already mounted: %s', share_path)
            return True
        else:
            LOG.debug('Share exists but is unavailable: %s ', share_path)
            if remove_unavailable_mapping:
                self.unmount_smb_share(share_path, force=True)
            return False

    def mount_smb_share(self, share_path, username=None, password=None):
        try:
            LOG.debug('Mounting share: %s', share_path)
            self._smb_conn.Msft_SmbMapping.Create(RemotePath=share_path,
                                                  UserName=username,
                                                  Password=password)
        except exceptions.x_wmi as exc:
            err_msg = (_(
                'Unable to mount SMBFS share: %(share_path)s '
                'WMI exception: %(wmi_exc)s') % {'share_path': share_path,
                                                 'wmi_exc': exc})
            raise exceptions.SMBException(err_msg)

    def unmount_smb_share(self, share_path, force=False):
        mappings = self._smb_conn.Msft_SmbMapping(RemotePath=share_path)
        if not mappings:
            LOG.debug('Share %s is not mounted. Skipping unmount.',
                      share_path)

        for mapping in mappings:
            # Due to a bug in the WMI module, getting the output of
            # methods returning None will raise an AttributeError
            try:
                mapping.Remove(Force=force)
            except AttributeError:
                pass
            except exceptions.x_wmi:
                # If this fails, a 'Generic Failure' exception is raised.
                # This happens even if we unforcefully unmount an in-use
                # share, for which reason we'll simply ignore it in this
                # case.
                if force:
                    raise exceptions.SMBException(
                        _("Could not unmount share: %s") % share_path)

    def get_smb_share_path(self, share_name):
        shares = self._smb_conn.Msft_SmbShare(Name=share_name)
        share_path = shares[0].Path if shares else None
        if not shares:
            LOG.debug("Could not find any local share named %s.", share_name)
        return share_path

    def is_local_share(self, share_path):
        # In case of Scale-Out File Servers, we'll get the Distributed Node
        # Name of the share. We have to check whether this resolves to a
        # local ip, which would happen in a hyper converged scenario.
        #
        # In this case, mounting the share is not supported and we have to
        # use the local share path.
        if share_path in self._loopback_share_map:
            return self._loopback_share_map[share_path]

        addr = share_path.lstrip('\\').split('\\', 1)[0]

        local_ips = _utils.get_ips(socket.gethostname())
        local_ips += _utils.get_ips('localhost')

        dest_ips = _utils.get_ips(addr)
        is_local = bool(set(local_ips).intersection(set(dest_ips)))

        self._loopback_share_map[share_path] = is_local
        return is_local



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

os_win_group = 'os_win'

os_win_opts = [
    cfg.StrOpt('hbaapi_lib_path',
               default='hbaapi.dll',
               help='Fibre Channel hbaapi library path. If no custom hbaapi '
                    'library is requested, the default one will be used.'),
    cfg.BoolOpt('cache_temporary_wmi_objects',
                default=True,
                help='Caches temporary WMI objects in order to increase '
                     'performance. This only affects networkutils, where '
                     'almost all operations require a reference to a '
                     'switch port. The cached objects are no longer valid '
                     'if the VM they are associated with is destroyed.'),
    cfg.IntOpt('wmi_job_terminate_timeout',
               default=120,
               help='The default amount of seconds to wait when stopping '
                    'pending WMI jobs. Setting this value to 0 will '
                    'disable the timeout.'),
]

CONF = cfg.CONF
CONF.register_opts(os_win_opts, os_win_group)


def list_opts():
    return [(os_win_group, os_win_opts)]



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes

from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi import wintypes

lib_handle = None

ISCSI_SECURITY_FLAGS = ctypes.c_uint64
ISCSI_LOGIN_FLAGS = ctypes.c_uint32
ISCSI_LOGIN_OPTIONS_INFO_SPECIFIED = ctypes.c_uint32
ISCSI_AUTH_TYPES = wintypes.UINT
ISCSI_DIGEST_TYPES = wintypes.UINT
DEVICE_TYPE = wintypes.ULONG


class ISCSI_TARGET_PORTAL(ctypes.Structure):
    _fields_ = [
        ('SymbolicName', wintypes.WCHAR * w_const.MAX_ISCSI_PORTAL_NAME_LEN),
        ('Address', wintypes.WCHAR * w_const.MAX_ISCSI_PORTAL_ADDRESS_LEN),
        ('Socket', wintypes.USHORT)
    ]


PISCSI_TARGET_PORTAL = ctypes.POINTER(ISCSI_TARGET_PORTAL)


class ISCSI_LOGIN_OPTIONS(ctypes.Structure):
    _fields_ = [
        ('Version', wintypes.ULONG),
        ('InformationSpecified', ISCSI_LOGIN_OPTIONS_INFO_SPECIFIED),
        ('LoginFlags', ISCSI_LOGIN_FLAGS),
        ('AuthType', ISCSI_AUTH_TYPES),
        ('HeaderDigest', ISCSI_DIGEST_TYPES),
        ('DataDigest', ISCSI_DIGEST_TYPES),
        ('MaximumConnections', wintypes.ULONG),
        ('DefaultTime2Wait', wintypes.ULONG),
        ('DefaultTime2Retain', wintypes.ULONG),
        ('UsernameLength', wintypes.ULONG),
        ('PasswordLength', wintypes.ULONG),
        ('Username', wintypes.PSTR),
        ('Password', wintypes.PSTR)
    ]


PISCSI_LOGIN_OPTIONS = ctypes.POINTER(ISCSI_LOGIN_OPTIONS)


class SCSI_LUN_LIST(ctypes.Structure):
    _fields_ = [
        ('OSLUN', wintypes.ULONG),
        ('TargetLUN', wintypes.ULONGLONG)
    ]


PSCSI_LUN_LIST = ctypes.POINTER(SCSI_LUN_LIST)


class ISCSI_UNIQUE_SESSION_ID(ctypes.Structure):
    _fields_ = [
        ('AdapterUnique', wintypes.ULONGLONG),
        ('AdapterSpecific', wintypes.ULONGLONG)
    ]


PISCSI_UNIQUE_SESSION_ID = ctypes.POINTER(ISCSI_UNIQUE_SESSION_ID)


class ISCSI_UNIQUE_CONNECTION_ID(ISCSI_UNIQUE_SESSION_ID):
    pass


PISCSI_UNIQUE_CONNECTION_ID = ctypes.POINTER(ISCSI_UNIQUE_CONNECTION_ID)


class ISCSI_TARGET_MAPPING(ctypes.Structure):
    _fields_ = [
        ('InitiatorName', wintypes.WCHAR * w_const.MAX_ISCSI_HBANAME_LEN),
        ('TargetName', wintypes.WCHAR * (w_const.MAX_ISCSI_NAME_LEN + 1)),
        ('OSDeviceName', wintypes.WCHAR * w_const.MAX_PATH),
        ('SessionId', ISCSI_UNIQUE_SESSION_ID),
        ('OSBusNumber', wintypes.ULONG),
        ('OSTargetNumber', wintypes.ULONG),
        ('LUNCount', wintypes.ULONG),
        ('LUNList', PSCSI_LUN_LIST)
    ]


PISCSI_TARGET_MAPPING = ctypes.POINTER(ISCSI_TARGET_MAPPING)


class PERSISTENT_ISCSI_LOGIN_INFO(ctypes.Structure):
    _fields_ = [
        ('TargetName', wintypes.WCHAR * (w_const.MAX_ISCSI_NAME_LEN + 1)),
        ('IsInformationalSession', wintypes.BOOLEAN),
        ('InitiatorInstance', wintypes.WCHAR * w_const.MAX_ISCSI_HBANAME_LEN),
        ('InitiatorPortNumber', wintypes.ULONG),
        ('TargetPortal', ISCSI_TARGET_PORTAL),
        ('SecurityFlags', ISCSI_SECURITY_FLAGS),
        ('Mappings', PISCSI_TARGET_MAPPING),
        ('LoginOptions', ISCSI_LOGIN_OPTIONS)
    ]


PPERSISTENT_ISCSI_LOGIN_INFO = ctypes.POINTER(PERSISTENT_ISCSI_LOGIN_INFO)


class ISCSI_CONNECTION_INFO(ctypes.Structure):
    _fields_ = [
        ('ConnectionId', ISCSI_UNIQUE_CONNECTION_ID),
        ('InitiatorAddress', wintypes.PWSTR),
        ('TargetAddress', wintypes.PWSTR),
        ('InitiatorSocket', wintypes.USHORT),
        ('TargetSocket', wintypes.USHORT),
        ('CID', ctypes.c_ubyte * 2)
    ]


PISCSI_CONNECTION_INFO = ctypes.POINTER(ISCSI_CONNECTION_INFO)


class ISCSI_SESSION_INFO(ctypes.Structure):
    _fields_ = [
        ('SessionId', ISCSI_UNIQUE_SESSION_ID),
        ('InitiatorName', wintypes.PWSTR),
        ('TargetName', wintypes.PWSTR),
        ('TargetNodeName', wintypes.PWSTR),
        ('ISID', ctypes.c_ubyte * 6),
        ('TSID', ctypes.c_ubyte * 2),
        ('ConnectionCount', wintypes.ULONG),
        ('Connections', PISCSI_CONNECTION_INFO)
    ]


PISCSI_SESSION_INFO = ctypes.POINTER(ISCSI_SESSION_INFO)


class SCSI_ADDRESS(ctypes.Structure):
    _fields_ = [
        ('Length', wintypes.ULONG),
        ('PortNumber', ctypes.c_ubyte),
        ('PathId', ctypes.c_ubyte),
        ('TargetId', ctypes.c_ubyte),
        ('Lun', ctypes.c_ubyte)
    ]


class STORAGE_DEVICE_NUMBER(ctypes.Structure):
    _fields_ = [
        ('DeviceType', DEVICE_TYPE),
        ('DeviceNumber', wintypes.DWORD),
        ('PartitionNumber', wintypes.DWORD)
    ]


class ISCSI_DEVICE_ON_SESSION(ctypes.Structure):
    _fields_ = [
        ('InitiatorName', wintypes.WCHAR * w_const.MAX_ISCSI_HBANAME_LEN),
        ('TargetName', wintypes.WCHAR * (w_const.MAX_ISCSI_NAME_LEN + 1)),
        ('ScsiAddress', SCSI_ADDRESS),
        ('DeviceInterfaceType', wintypes.GUID),
        ('DeviceInterfaceName', wintypes.WCHAR * w_const.MAX_PATH),
        ('LegacyName', wintypes.WCHAR * w_const.MAX_PATH),
        ('StorageDeviceNumber', STORAGE_DEVICE_NUMBER),
        ('DeviceInstance', wintypes.ULONG)
    ]


PISCSI_DEVICE_ON_SESSION = ctypes.POINTER(ISCSI_DEVICE_ON_SESSION)


def register():
    global lib_handle
    lib_handle = ctypes.windll.iscsidsc

    lib_handle.AddIScsiStaticTargetW.argtypes = [
        wintypes.PWSTR,
        wintypes.PWSTR,
        wintypes.ULONG,
        wintypes.BOOLEAN,
        PISCSI_TARGET_MAPPING,
        PISCSI_LOGIN_OPTIONS,
        wintypes.LPVOID  # unused
    ]
    lib_handle.AddIScsiStaticTargetW.restype = wintypes.ULONG

    lib_handle.GetDevicesForIScsiSessionW.argtypes = [
        PISCSI_UNIQUE_SESSION_ID,
        wintypes.PULONG,
        PISCSI_DEVICE_ON_SESSION
    ]
    lib_handle.GetDevicesForIScsiSessionW.restype = wintypes.ULONG

    lib_handle.GetIScsiInitiatorNodeNameW.argtypes = [wintypes.PWCHAR]
    lib_handle.GetIScsiInitiatorNodeNameW.restype = wintypes.ULONG

    lib_handle.GetIScsiSessionListW.argtypes = [
        wintypes.PULONG,
        wintypes.PULONG,
        PISCSI_SESSION_INFO
    ]
    lib_handle.GetIScsiSessionListW.restype = wintypes.ULONG

    lib_handle.LoginIScsiTargetW.argtypes = [
        wintypes.PWSTR,
        wintypes.BOOLEAN,
        wintypes.PWSTR,
        wintypes.ULONG,
        PISCSI_TARGET_PORTAL,
        ISCSI_SECURITY_FLAGS,
        PISCSI_TARGET_MAPPING,
        PISCSI_LOGIN_OPTIONS,
        wintypes.ULONG,
        wintypes.PCHAR,
        wintypes.BOOLEAN,
        PISCSI_UNIQUE_SESSION_ID,
        PISCSI_UNIQUE_CONNECTION_ID
    ]
    lib_handle.LoginIScsiTargetW.restype = wintypes.ULONG

    lib_handle.LogoutIScsiTarget.argtypes = [PISCSI_UNIQUE_SESSION_ID]
    lib_handle.LogoutIScsiTarget.restype = wintypes.ULONG

    lib_handle.RemoveIScsiPersistentTargetW.argtypes = [
        wintypes.PWSTR,
        wintypes.ULONG,
        wintypes.PWSTR,
        PISCSI_TARGET_PORTAL
    ]
    lib_handle.RemoveIScsiPersistentTargetW.restype = wintypes.ULONG

    lib_handle.RemoveIScsiStaticTargetW.argtypes = [wintypes.PWSTR]
    lib_handle.RemoveIScsiStaticTargetW.restype = wintypes.ULONG

    lib_handle.ReportIScsiInitiatorListW.argtypes = [
        wintypes.PULONG,
        wintypes.PWCHAR
    ]
    lib_handle.ReportIScsiInitiatorListW.restype = wintypes.ULONG

    lib_handle.ReportIScsiPersistentLoginsW.argtypes = [
        wintypes.PULONG,
        PPERSISTENT_ISCSI_LOGIN_INFO,
        wintypes.PULONG
    ]
    lib_handle.ReportIScsiPersistentLoginsW.restype = wintypes.ULONG

    lib_handle.ReportIScsiTargetsW.argtypes = [
        wintypes.BOOLEAN,
        wintypes.PULONG,
        wintypes.PWCHAR
    ]
    lib_handle.ReportIScsiTargetsW.restype = wintypes.ULONG



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Utility class for VM related operations on Hyper-V Clusters.
"""

import ctypes
import re
import sys
import threading
import time

from eventlet import patcher
from eventlet import tpool
from oslo_log import log as logging
from oslo_utils import excutils
from six.moves import queue

from os_win._i18n import _
from os_win import _utils
from os_win import constants
from os_win import exceptions
from os_win.utils import baseutils
from os_win.utils.compute import _clusapi_utils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi.libs import clusapi as clusapi_def
from os_win.utils.winapi import wintypes

LOG = logging.getLogger(__name__)


class ClusterUtils(baseutils.BaseUtils):

    _MSCLUSTER_NODE = 'MSCluster_Node'
    _MSCLUSTER_RES = 'MSCluster_Resource'

    _VM_BASE_NAME = 'Virtual Machine %s'
    _VM_TYPE = 'Virtual Machine'

    _MS_CLUSTER_NAMESPACE = '//%s/root/MSCluster'

    _LIVE_MIGRATION_TYPE = 4
    _IGNORE_LOCKED = 1
    _DESTROY_GROUP = 1

    _FAILBACK_WINDOW_MIN = 0
    _FAILBACK_WINDOW_MAX = 23

    _WMI_EVENT_TIMEOUT_MS = 100
    _WMI_EVENT_CHECK_INTERVAL = 2

    def __init__(self, host='.'):
        self._instance_name_regex = re.compile('Virtual Machine (.*)')
        self._clusapi_utils = _clusapi_utils.ClusApiUtils()
        self._cmgr = _clusapi_utils.ClusterContextManager()

        if sys.platform == 'win32':
            self._init_hyperv_conn(host)

    def _init_hyperv_conn(self, host):
        try:
            self._conn_cluster = self._get_wmi_conn(
                self._MS_CLUSTER_NAMESPACE % host)
            self._cluster = self._conn_cluster.MSCluster_Cluster()[0]

            # extract this node name from cluster's path
            path = self._cluster.path_()
            self._this_node = re.search(r'\\\\(.*)\\root', path,
                                        re.IGNORECASE).group(1)
        except AttributeError:
            raise exceptions.HyperVClusterException(
                _("Could not initialize cluster wmi connection."))

    def _get_failover_watcher(self):
        raw_query = ("SELECT * FROM __InstanceModificationEvent "
                     "WITHIN %(wmi_check_interv)s WHERE TargetInstance ISA "
                     "'%(cluster_res)s' AND "
                     "TargetInstance.Type='%(cluster_res_type)s' AND "
                     "TargetInstance.OwnerNode != PreviousInstance.OwnerNode" %
                     {'wmi_check_interv': self._WMI_EVENT_CHECK_INTERVAL,
                      'cluster_res': self._MSCLUSTER_RES,
                      'cluster_res_type': self._VM_TYPE})
        return self._conn_cluster.watch_for(raw_wql=raw_query)

    def check_cluster_state(self):
        if len(list(self._get_cluster_nodes())) < 1:
            raise exceptions.HyperVClusterException(
                _("Not enough cluster nodes."))

    def get_node_name(self):
        return self._this_node

    def _get_cluster_nodes(self):
        return self.cluster_enum(w_const.CLUSTER_ENUM_NODE)

    def _get_vm_groups(self):
        for r in self.cluster_enum(w_const.CLUSTER_ENUM_GROUP):
            group_type = self.get_cluster_group_type(r['name'])
            if group_type == w_const.ClusGroupTypeVirtualMachine:
                yield r

    def _lookup_vm_group_check(self, vm_name):
        vm = self._lookup_vm_group(vm_name)
        if not vm:
            raise exceptions.HyperVVMNotFoundException(vm_name=vm_name)
        return vm

    def _lookup_vm_group(self, vm_name):
        return self._lookup_res(self._conn_cluster.MSCluster_ResourceGroup,
                                vm_name)

    def _lookup_res(self, resource_source, res_name):
        res = resource_source(Name=res_name)
        n = len(res)
        if n == 0:
            return None
        elif n > 1:
            raise exceptions.HyperVClusterException(
                _('Duplicate resource name %s found.') % res_name)
        else:
            return res[0]

    def get_cluster_node_names(self):
        nodes = self._get_cluster_nodes()
        return [n['name'] for n in nodes]

    def get_vm_host(self, vm_name):
        with self._cmgr.open_cluster_group(vm_name) as group_handle:
            state_info = self._get_cluster_group_state(group_handle)
            return state_info['owner_node']

    def list_instances(self):
        return [r['name'] for r in self._get_vm_groups()]

    def list_instance_uuids(self):
        return [r['id'] for r in self._get_vm_groups()]

    def add_vm_to_cluster(self, vm_name, max_failover_count=1,
                          failover_period=6, auto_failback=True):
        """Adds the VM to the Hyper-V Cluster.

        :param vm_name: The name of the VM to be added to the Hyper-V Cluster
        :param max_failover_count: The number of times the Hyper-V Cluster will
            try to failover the VM within the given failover period. If the VM
            will try to failover more than this number of the given
            failover_period, the VM will end up in a failed state.
        :param failover_period: The period (hours) over which the given
            max_failover_count failovers can occur. After this period expired,
            the failover count for the given VM is reset.
        :param auto_failback: boolean, whether the VM will be allowed to
            move back to its original host when it is available again.
        """
        LOG.debug("Add vm to cluster called for vm %s" % vm_name)
        self._cluster.AddVirtualMachine(vm_name)

        vm_group = self._lookup_vm_group_check(vm_name)
        vm_group.FailoverThreshold = max_failover_count
        vm_group.FailoverPeriod = failover_period
        vm_group.PersistentState = True
        vm_group.AutoFailbackType = int(bool(auto_failback))
        # set the earliest and latest time that the group can be moved
        # back to its preferred node. The unit is in hours.
        vm_group.FailbackWindowStart = self._FAILBACK_WINDOW_MIN
        vm_group.FailbackWindowEnd = self._FAILBACK_WINDOW_MAX
        vm_group.put()

    def bring_online(self, vm_name):
        with self._cmgr.open_cluster_group(vm_name) as group_handle:
            self._clusapi_utils.online_cluster_group(group_handle)

    def take_offline(self, vm_name):
        with self._cmgr.open_cluster_group(vm_name) as group_handle:
            self._clusapi_utils.offline_cluster_group(group_handle)

    def delete(self, vm_name):
        # We're sticking with WMI, for now. Destroying VM cluster groups using
        # clusapi's DestroyClusterGroup function acts strange. VMs get
        # recreated asyncronuously and put in suspended state,
        # breaking everything.
        vm = self._lookup_vm_group_check(vm_name)
        vm.DestroyGroup(self._DESTROY_GROUP)

    def cluster_enum(self, object_type):
        with self._cmgr.open_cluster_enum(object_type) as enum_handle:
            object_count = self._clusapi_utils.cluster_get_enum_count(
                enum_handle)
            for idx in range(object_count):
                item = self._clusapi_utils.cluster_enum(enum_handle, idx)

                item_dict = dict(version=item.dwVersion,
                                 type=item.dwType,
                                 id=item.lpszId,
                                 name=item.lpszName)
                yield item_dict

    def vm_exists(self, vm_name):
        res_name = self._VM_BASE_NAME % vm_name
        try:
            with self._cmgr.open_cluster_resource(res_name):
                return True
        except exceptions.ClusterObjectNotFound:
            return False

    def live_migrate_vm(self, vm_name, new_host, timeout=None):
        self._migrate_vm(vm_name, new_host, self._LIVE_MIGRATION_TYPE,
                         constants.CLUSTER_GROUP_ONLINE,
                         timeout)

    def _migrate_vm(self, vm_name, new_host, migration_type,
                    exp_state_after_migr, timeout):
        syntax = w_const.CLUSPROP_SYNTAX_LIST_VALUE_DWORD
        migr_type = wintypes.DWORD(migration_type)

        prop_entries = [
            self._clusapi_utils.get_property_list_entry(
                w_const.CLUS_RESTYPE_NAME_VM, syntax, migr_type),
            self._clusapi_utils.get_property_list_entry(
                w_const.CLUS_RESTYPE_NAME_VM_CONFIG, syntax, migr_type)
        ]
        prop_list = self._clusapi_utils.get_property_list(prop_entries)

        flags = (
            w_const.CLUSAPI_GROUP_MOVE_RETURN_TO_SOURCE_NODE_ON_ERROR |
            w_const.CLUSAPI_GROUP_MOVE_QUEUE_ENABLED |
            w_const.CLUSAPI_GROUP_MOVE_HIGH_PRIORITY_START)

        with self._cmgr.open_cluster() as cluster_handle, \
                self._cmgr.open_cluster_group(
                    vm_name,
                    cluster_handle=cluster_handle) as group_handle, \
                self._cmgr.open_cluster_node(
                    new_host,
                    cluster_handle=cluster_handle) as dest_node_handle, \
                _ClusterGroupStateChangeListener(cluster_handle,
                                                 vm_name) as listener:
            self._clusapi_utils.move_cluster_group(group_handle,
                                                   dest_node_handle,
                                                   flags,
                                                   prop_list)
            try:
                self._wait_for_cluster_group_migration(
                    listener,
                    vm_name,
                    group_handle,
                    exp_state_after_migr,
                    timeout)
            except exceptions.ClusterGroupMigrationTimeOut:
                with excutils.save_and_reraise_exception() as ctxt:
                    self._cancel_cluster_group_migration(
                        listener, vm_name, group_handle,
                        exp_state_after_migr, timeout)

                    # This is rather unlikely to happen but we're
                    # covering it out.
                    try:
                        self._validate_migration(group_handle,
                                                 vm_name,
                                                 exp_state_after_migr,
                                                 new_host)
                        LOG.warning(
                            'Cluster group migration completed '
                            'successfuly after cancel attempt. '
                            'Suppressing timeout exception.')
                        ctxt.reraise = False
                    except exceptions.ClusterGroupMigrationFailed:
                        pass
            else:
                self._validate_migration(group_handle,
                                         vm_name,
                                         exp_state_after_migr,
                                         new_host)

    def _validate_migration(self, group_handle, group_name,
                            expected_state, expected_node):
        state_info = self._clusapi_utils.get_cluster_group_state(group_handle)
        owner_node = state_info['owner_node']
        group_state = state_info['state']

        if (expected_state != group_state or
                expected_node.lower() != owner_node.lower()):
            raise exceptions.ClusterGroupMigrationFailed(
                group_name=group_name,
                expected_state=expected_state,
                expected_node=expected_node,
                group_state=group_state,
                owner_node=owner_node)

    def cancel_cluster_group_migration(self, group_name, expected_state,
                                       timeout=None):
        with self._cmgr.open_cluster() as cluster_handle, \
                self._cmgr.open_cluster_group(
                    group_name,
                    cluster_handle=cluster_handle) as group_handle, \
                _ClusterGroupStateChangeListener(cluster_handle,
                                                 group_name) as listener:
            self._cancel_cluster_group_migration(
                listener, group_name, group_handle,
                expected_state, timeout)

    def _cancel_cluster_group_migration(self, event_listener,
                                        group_name, group_handle,
                                        expected_state,
                                        timeout=None):
        LOG.info("Canceling cluster group '%s' migration", group_name)
        try:
            cancel_finished = (
                self._clusapi_utils.cancel_cluster_group_operation(
                    group_handle))
        except exceptions.Win32Exception as ex:
            group_state_info = self._get_cluster_group_state(group_handle)
            migration_pending = self._is_migration_pending(
                group_state_info['state'],
                group_state_info['status_info'],
                expected_state)

            if (ex.error_code == w_const.ERROR_INVALID_STATE and
                    not migration_pending):
                LOG.debug('Ignoring group migration cancel error. '
                          'No migration is pending.')
                cancel_finished = True
            else:
                raise

        if not cancel_finished:
            LOG.debug("Waiting for group migration to be canceled.")
            try:
                self._wait_for_cluster_group_migration(
                    event_listener, group_name, group_handle,
                    expected_state,
                    timeout=timeout)
            except Exception:
                LOG.exception("Failed to cancel cluster group migration.")
                raise exceptions.JobTerminateFailed()

        LOG.info("Cluster group migration canceled.")

    def _is_migration_queued(self, group_status_info):
        return bool(
            group_status_info &
            w_const.CLUSGRP_STATUS_WAITING_IN_QUEUE_FOR_MOVE)

    def _is_migration_pending(self, group_state, group_status_info,
                              expected_state):
        migration_pending = (
            group_state != expected_state or
            self._is_migration_queued(group_status_info))
        return migration_pending

    def _wait_for_cluster_group_migration(self, event_listener,
                                          group_name, group_handle,
                                          expected_state,
                                          timeout=None):
        time_start = time.time()
        time_left = timeout if timeout else 'undefined'

        group_state_info = self._get_cluster_group_state(group_handle)
        group_state = group_state_info['state']
        group_status_info = group_state_info['status_info']

        migration_pending = self._is_migration_pending(
            group_state,
            group_status_info,
            expected_state)
        if not migration_pending:
            return

        while not timeout or time_left > 0:
            time_elapsed = time.time() - time_start
            time_left = timeout - time_elapsed if timeout else 'undefined'

            LOG.debug("Waiting for cluster group '%(group_name)s' "
                      "migration to finish. "
                      "Time left: %(time_left)s.",
                      dict(group_name=group_name,
                           time_left=time_left))

            try:
                event = event_listener.get(time_left if timeout else None)
            except queue.Empty:
                break

            group_state = event.get('state', group_state)
            group_status_info = event.get('status_info', group_status_info)

            migration_pending = self._is_migration_pending(group_state,
                                                           group_status_info,
                                                           expected_state)
            if not migration_pending:
                return

        LOG.error("Cluster group migration timed out.")
        raise exceptions.ClusterGroupMigrationTimeOut(
            group_name=group_name,
            time_elapsed=time.time() - time_start)

    def get_cluster_node_name(self, node_id):
        for node in self._get_cluster_nodes():
            if node['id'] == node_id:
                return node['name']

        err_msg = _("Could not find any cluster node with id: %s.")
        raise exceptions.NotFound(err_msg % node_id)

    def get_cluster_group_type(self, group_name):
        with self._cmgr.open_cluster_group(group_name) as group_handle:
            buff, buff_sz = self._clusapi_utils.cluster_group_control(
                group_handle, w_const.CLUSCTL_GROUP_GET_RO_COMMON_PROPERTIES)
            return self._clusapi_utils.get_cluster_group_type(
                ctypes.byref(buff), buff_sz)

    def get_cluster_group_state_info(self, group_name):
        """Gets cluster group state info.

        :return: a dict containing the following keys:
            ['state', 'migration_queued', 'owner_node']
        """
        with self._cmgr.open_cluster_group(group_name) as group_handle:
            state_info = self._get_cluster_group_state(group_handle)
            migration_queued = self._is_migration_queued(
                state_info['status_info'])

            return dict(owner_node=state_info['owner_node'],
                        state=state_info['state'],
                        migration_queued=migration_queued)

    def _get_cluster_group_state(self, group_handle):
        state_info = self._clusapi_utils.get_cluster_group_state(group_handle)

        buff, buff_sz = self._clusapi_utils.cluster_group_control(
            group_handle,
            w_const.CLUSCTL_GROUP_GET_RO_COMMON_PROPERTIES)
        status_info = self._clusapi_utils.get_cluster_group_status_info(
            ctypes.byref(buff), buff_sz)

        state_info['status_info'] = status_info
        return state_info

    def _monitor_vm_failover(self, watcher, callback,
                             event_timeout_ms=_WMI_EVENT_TIMEOUT_MS):
        """Creates a monitor to check for new WMI MSCluster_Resource

        events.

        This method will poll the last _WMI_EVENT_CHECK_INTERVAL + 1
        seconds for new events and listens for _WMI_EVENT_TIMEOUT_MS
        milliseconds, since listening is a thread blocking action.

        Any event object caught will then be processed.
        """

        vm_name = None
        new_host = None
        try:
            # wait for new event for _WMI_EVENT_TIMEOUT_MS milliseconds.
            if patcher.is_monkey_patched('thread'):
                wmi_object = tpool.execute(watcher,
                                           event_timeout_ms)
            else:
                wmi_object = watcher(event_timeout_ms)

            old_host = wmi_object.previous.OwnerNode
            new_host = wmi_object.OwnerNode
            # wmi_object.Name field is of the form:
            # 'Virtual Machine nova-instance-template'
            # wmi_object.Name filed is a key and as such is not affected
            # by locale, so it will always be 'Virtual Machine'
            match = self._instance_name_regex.search(wmi_object.Name)
            if match:
                vm_name = match.group(1)

            if vm_name:
                try:
                    callback(vm_name, old_host, new_host)
                except Exception:
                    LOG.exception(
                        "Exception during failover callback.")
        except exceptions.x_wmi_timed_out:
            pass

    def get_vm_owner_change_listener(self):
        def listener(callback):
            watcher = self._get_failover_watcher()

            while True:
                # We avoid setting an infinite timeout in order to let
                # the process gracefully stop. Note that the os-win WMI
                # event listeners are meant to be used as long running
                # daemons, so no stop API is provided ATM.
                try:
                    self._monitor_vm_failover(
                        watcher,
                        callback,
                        constants.DEFAULT_WMI_EVENT_TIMEOUT_MS)
                except Exception:
                    LOG.exception("The VM cluster group owner change "
                                  "event listener encountered an "
                                  "unexpected exception.")
                    time.sleep(constants.DEFAULT_WMI_EVENT_TIMEOUT_MS / 1000)

        return listener

    def get_vm_owner_change_listener_v2(self):
        def listener(callback):
            cluster_handle = self._clusapi_utils.open_cluster()
            _listener = _ClusterGroupOwnerChangeListener(cluster_handle)

            while True:
                try:
                    event = _listener.get()
                    group_name = event['cluster_object_name']
                    group_type = self.get_cluster_group_type(group_name)
                    if group_type != w_const.ClusGroupTypeVirtualMachine:
                        continue

                    new_node_id = event['parent_id']
                    new_node_name = self.get_cluster_node_name(new_node_id)
                    callback(group_name, new_node_name)
                except Exception:
                    LOG.exception("The VM cluster group owner change "
                                  "event listener encountered an "
                                  "unexpected exception.")
                    time.sleep(constants.DEFAULT_WMI_EVENT_TIMEOUT_MS / 1000)

        return listener


# At the moment, those event listeners are not meant to be used outside
# os-win, mostly because of the underlying API limitations.
class _ClusterEventListener(object):
    _notif_keys = {}
    _notif_port_h = None
    _cluster_handle = None
    _running = False
    _stop_on_error = True
    _error_sleep_interval = 2

    def __init__(self, cluster_handle, stop_on_error=True):
        self._cluster_handle = cluster_handle
        self._stop_on_error = stop_on_error

        self._clusapi_utils = _clusapi_utils.ClusApiUtils()
        self._event_queue = queue.Queue()

        self._setup()

    def __enter__(self):
        self._ensure_listener_running()
        return self

    def _get_notif_key_dw(self, notif_key):
        notif_key_dw = self._notif_keys.get(notif_key)
        if notif_key_dw is None:
            notif_key_dw = wintypes.DWORD(notif_key)
            # We have to make sure those addresses are preserved.
            self._notif_keys[notif_key] = notif_key_dw
        return notif_key_dw

    def _add_filter(self, notif_filter, notif_key=0):
        notif_key_dw = self._get_notif_key_dw(notif_key)

        # We'll get a notification handle if not already existing.
        self._notif_port_h = self._clusapi_utils.create_cluster_notify_port_v2(
            self._cluster_handle, notif_filter,
            self._notif_port_h, notif_key_dw)

    def _setup_notif_port(self):
        for notif_filter in self._notif_filters_list:
            filter_struct = clusapi_def.NOTIFY_FILTER_AND_TYPE(
                dwObjectType=notif_filter['object_type'],
                FilterFlags=notif_filter['filter_flags'])
            notif_key = notif_filter.get('notif_key', 0)

            self._add_filter(filter_struct, notif_key)

    def _setup(self):
        self._setup_notif_port()

        # If eventlet monkey patching is used, this will actually be a
        # greenthread. We just don't want to enforce eventlet usage.
        worker = threading.Thread(target=self._listen)
        worker.setDaemon(True)

        self._running = True
        worker.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def _signal_stopped(self):
        self._running = False
        self._event_queue.put(None)

    def stop(self):
        self._signal_stopped()

        if self._notif_port_h:
            self._clusapi_utils.close_cluster_notify_port(self._notif_port_h)

    def _listen(self):
        while self._running:
            try:
                # We're using an indefinite timeout here. When the listener is
                # closed, this will raise an 'invalid handle value' error,
                # which we're going to ignore.
                event = _utils.avoid_blocking_call(
                    self._clusapi_utils.get_cluster_notify_v2,
                    self._notif_port_h,
                    timeout_ms=-1)

                processed_event = self._process_event(event)
                if processed_event:
                    self._event_queue.put(processed_event)
            except Exception:
                if self._running:
                    LOG.exception(
                        "Unexpected exception in event listener loop.")
                    if self._stop_on_error:
                        LOG.warning(
                            "The cluster event listener will now close.")
                        self._signal_stopped()
                    else:
                        time.sleep(self._error_sleep_interval)

    def _process_event(self, event):
        return event

    def get(self, timeout=None):
        self._ensure_listener_running()

        event = self._event_queue.get(timeout=timeout)

        self._ensure_listener_running()
        return event

    def _ensure_listener_running(self):
        if not self._running:
            raise exceptions.OSWinException(
                _("Cluster event listener is not running."))


class _ClusterGroupStateChangeListener(_ClusterEventListener):
    _NOTIF_KEY_GROUP_STATE = 0
    _NOTIF_KEY_GROUP_COMMON_PROP = 1

    _notif_filters_list = [
        dict(object_type=w_const.CLUSTER_OBJECT_TYPE_GROUP,
             filter_flags=w_const.CLUSTER_CHANGE_GROUP_STATE_V2,
             notif_key=_NOTIF_KEY_GROUP_STATE),
        dict(object_type=w_const.CLUSTER_OBJECT_TYPE_GROUP,
             filter_flags=w_const.CLUSTER_CHANGE_GROUP_COMMON_PROPERTY_V2,
             notif_key=_NOTIF_KEY_GROUP_COMMON_PROP)]

    def __init__(self, cluster_handle, group_name=None, **kwargs):
        self._group_name = group_name

        super(_ClusterGroupStateChangeListener, self).__init__(
            cluster_handle, **kwargs)

    def _process_event(self, event):
        group_name = event['cluster_object_name']
        if self._group_name and self._group_name.lower() != group_name.lower():
            return

        preserved_keys = ['cluster_object_name', 'object_type',
                          'filter_flags', 'notif_key']
        processed_event = {key: event[key] for key in preserved_keys}

        notif_key = event['notif_key']
        if notif_key == self._NOTIF_KEY_GROUP_STATE:
            if event['buff_sz'] != ctypes.sizeof(wintypes.DWORD):
                raise exceptions.ClusterPropertyRetrieveFailed()
            state_p = ctypes.cast(event['buff'], wintypes.PDWORD)
            state = state_p.contents.value
            processed_event['state'] = state
            return processed_event
        elif notif_key == self._NOTIF_KEY_GROUP_COMMON_PROP:
            try:
                status_info = (
                    self._clusapi_utils.get_cluster_group_status_info(
                        ctypes.byref(event['buff']), event['buff_sz']))
                processed_event['status_info'] = status_info
                return processed_event
            except exceptions.ClusterPropertyListEntryNotFound:
                # At the moment, we only care about the 'StatusInformation'
                # common property.
                pass


class _ClusterGroupOwnerChangeListener(_ClusterEventListener):
    _notif_filters_list = [
        dict(object_type=w_const.CLUSTER_OBJECT_TYPE_GROUP,
             filter_flags=w_const.CLUSTER_CHANGE_GROUP_OWNER_NODE_V2)
    ]



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib

from os_win._i18n import _
from os_win import exceptions


ADVAPI32 = 'advapi32'
CLUSAPI = 'clusapi'
HBAAPI = 'hbaapi'
ISCSIDSC = 'iscsidsc'
KERNEL32 = 'kernel32'
VIRTDISK = 'virtdisk'

libs = [ADVAPI32, CLUSAPI, HBAAPI, ISCSIDSC, KERNEL32, VIRTDISK]


def _get_shared_lib_module(lib_name):
    if lib_name not in libs:
        err_msg = _("Unsupported library: %s.")
        raise exceptions.OSWinException(err_msg % lib_name)

    module = importlib.import_module('os_win.utils.winapi.libs.%s' % lib_name)
    return module


def register():
    for lib_name in libs:
        module = _get_shared_lib_module(lib_name)
        module.register()


def get_shared_lib_handle(lib_name):
    module = _get_shared_lib_module(lib_name)
    return module.lib_handle



# Copyright 2013 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket

from oslo_log import log as logging

from os_win._i18n import _
from os_win import _utils
from os_win import constants
from os_win import exceptions
from os_win.utils import baseutils
from os_win.utils.winapi import libs as w_lib

kernel32 = w_lib.get_shared_lib_handle(w_lib.KERNEL32)

LOG = logging.getLogger(__name__)


class HostUtils(baseutils.BaseUtilsVirt):

    _windows_version = None

    _MSVM_PROCESSOR = 'Msvm_Processor'
    _MSVM_MEMORY = 'Msvm_Memory'
    _MSVM_NUMA_NODE = 'Msvm_NumaNode'

    _CENTRAL_PROCESSOR = 'Central Processor'

    _HOST_FORCED_REBOOT = 6
    _HOST_FORCED_SHUTDOWN = 12
    _DEFAULT_VM_GENERATION = constants.IMAGE_PROP_VM_GEN_1

    FEATURE_RDS_VIRTUALIZATION = 322
    FEATURE_MPIO = 57

    _wmi_cimv2_namespace = '//./root/cimv2'
    _wmi_standard_cimv2_namespace = '//./root/StandardCimv2'

    def __init__(self, host='.'):
        super(HostUtils, self).__init__(host)
        self._conn_cimv2 = self._get_wmi_conn(self._wmi_cimv2_namespace,
                                              privileges=["Shutdown"])
        self._conn_scimv2 = self._get_wmi_conn(
            self._wmi_standard_cimv2_namespace)
        self._netutils_prop = None

    @property
    def _netutils(self):
        if not self._netutils_prop:
            # NOTE(claudiub): we're importing utilsfactory here in order to
            # avoid circular dependencies.
            from os_win import utilsfactory
            self._netutils_prop = utilsfactory.get_networkutils()

        return self._netutils_prop

    def get_cpus_info(self):
        """Returns dictionary containing information about the host's CPUs."""
        # NOTE(abalutoiu): Specifying exactly the fields that we need
        # improves the speed of the query. The LoadPercentage field
        # is the load capacity of each processor averaged to the last
        # second, which is time wasted.
        cpus = self._conn_cimv2.query(
            "SELECT Architecture, Name, Manufacturer, MaxClockSpeed, "
            "NumberOfCores, NumberOfLogicalProcessors FROM Win32_Processor "
            "WHERE ProcessorType = 3")
        cpus_list = []
        for cpu in cpus:
            cpu_info = {'Architecture': cpu.Architecture,
                        'Name': cpu.Name,
                        'Manufacturer': cpu.Manufacturer,
                        'MaxClockSpeed': cpu.MaxClockSpeed,
                        'NumberOfCores': cpu.NumberOfCores,
                        'NumberOfLogicalProcessors':
                        cpu.NumberOfLogicalProcessors}
            cpus_list.append(cpu_info)
        return cpus_list

    def is_cpu_feature_present(self, feature_key):
        """Checks if the host's CPUs have the given feature."""
        return kernel32.IsProcessorFeaturePresent(feature_key)

    def get_memory_info(self):
        """Returns a tuple with total visible memory and free physical memory.

        The returned values are expressed in KB.
        """

        mem_info = self._conn_cimv2.query("SELECT TotalVisibleMemorySize, "
                                          "FreePhysicalMemory "
                                          "FROM win32_operatingsystem")[0]
        return (int(mem_info.TotalVisibleMemorySize),
                int(mem_info.FreePhysicalMemory))

    # TODO(atuvenie) This method should be removed once all the callers have
    # changed to use the get_disk_capacity method from diskutils.
    def get_volume_info(self, drive):
        """Returns a tuple with total size and free space of the given drive.

        Returned values are expressed in bytes.

        :param drive: the drive letter of the logical disk whose information
            is required.
        """

        logical_disk = self._conn_cimv2.query("SELECT Size, FreeSpace "
                                              "FROM win32_logicaldisk "
                                              "WHERE DeviceID='%s'"
                                              % drive)[0]
        return (int(logical_disk.Size), int(logical_disk.FreeSpace))

    def check_min_windows_version(self, major, minor, build=0):
        """Compares the host's kernel version with the given version.

        :returns: True if the host's kernel version is higher or equal to
            the given version.
        """
        version_str = self.get_windows_version()
        return list(map(int, version_str.split('.'))) >= [major, minor, build]

    def get_windows_version(self):
        """Returns a string representing the host's kernel version."""
        if not HostUtils._windows_version:
            Win32_OperatingSystem = self._conn_cimv2.Win32_OperatingSystem()[0]
            HostUtils._windows_version = Win32_OperatingSystem.Version
        return HostUtils._windows_version

    def get_local_ips(self):
        """Returns the list of locally assigned IPs."""
        hostname = socket.gethostname()
        return _utils.get_ips(hostname)

    def get_host_tick_count64(self):
        """Returns host uptime in milliseconds."""
        return kernel32.GetTickCount64()

    def host_power_action(self, action):
        win32_os = self._conn_cimv2.Win32_OperatingSystem()[0]

        if action == constants.HOST_POWER_ACTION_SHUTDOWN:
            win32_os.Win32Shutdown(self._HOST_FORCED_SHUTDOWN)
        elif action == constants.HOST_POWER_ACTION_REBOOT:
            win32_os.Win32Shutdown(self._HOST_FORCED_REBOOT)
        else:
            raise NotImplementedError(
                _("Host %(action)s is not supported by the Hyper-V driver") %
                {"action": action})

    def get_supported_vm_types(self):
        """Get the supported Hyper-V VM generations.

        Hyper-V Generation 2 VMs are supported in Windows 8.1,
        Windows Server / Hyper-V Server 2012 R2 or newer.

        :returns: array of supported VM generations (ex. ['hyperv-gen1'])
        """

        if self.check_min_windows_version(6, 3):
            return [constants.IMAGE_PROP_VM_GEN_1,
                    constants.IMAGE_PROP_VM_GEN_2]
        else:
            return [constants.IMAGE_PROP_VM_GEN_1]

    def get_default_vm_generation(self):
        return self._DEFAULT_VM_GENERATION

    def check_server_feature(self, feature_id):
        """Checks if the given feature exists on the host."""
        return len(self._conn_cimv2.Win32_ServerFeature(ID=feature_id)) > 0

    def get_nic_sriov_vfs(self):
        """Get host's NIC SR-IOV VFs.

        This method will ignore the vSwitches which do not have SR-IOV enabled,
        or which are poorly configured (the NIC does not support SR-IOV).

        :returns: a list of dictionaries, containing the following fields:
            - 'vswitch_name': the vSwtch name.
            - 'total_vfs': the vSwitch's maximum number of VFs. (> 0)
            - 'used_vfs': the vSwitch's number of used VFs. (<= 'total_vfs')
        """

        # TODO(claudiub): We have added a different method that returns all
        # of the offloading capabilities available, including SR-IOV.
        # Remove this method in S.

        vfs = []

        # NOTE(claudiub): A vSwitch will have to be configured to enable
        # SR-IOV, otherwise its IOVPreferred flag will be False.
        vswitch_sds = self._conn.Msvm_VirtualEthernetSwitchSettingData(
            IOVPreferred=True)
        for vswitch_sd in vswitch_sds:
            hw_offload = self._conn.Msvm_EthernetSwitchHardwareOffloadData(
                SystemName=vswitch_sd.VirtualSystemIdentifier)[0]
            if not hw_offload.IovVfCapacity:
                LOG.warning("VSwitch %s has SR-IOV enabled, but it is not "
                            "supported by the NIC or by the OS.",
                            vswitch_sd.ElementName)
                continue

            nic_name = self._netutils.get_vswitch_external_network_name(
                vswitch_sd.ElementName)
            if not nic_name:
                # NOTE(claudiub): This can happen if the vSwitch is not
                # external.
                LOG.warning("VSwitch %s is not external.",
                            vswitch_sd.ElementName)
                continue

            nic = self._conn_scimv2.MSFT_NetAdapter(
                InterfaceDescription=nic_name)[0]

            vfs.append({
                'vswitch_name': vswitch_sd.ElementName,
                'device_id': nic.PnPDeviceID,
                'total_vfs': hw_offload.IovVfCapacity,
                'used_vfs': hw_offload.IovVfUsage,
            })

        return vfs

    def get_nic_hardware_offload_info(self):
        """Get host's NIC hardware offload information.

        Hyper-V offers a few different hardware offloading options for VMs and
        their vNICs, depending on the vSwitches' NICs hardware resources and
        capabilities. These resources are managed and assigned automatically by
        Hyper-V. These resources are: VFs, IOV queue pairs, VMQs, IPsec
        security association offloads.

        :returns: a list of dictionaries, containing the following fields:
            - 'vswitch_name': the switch name.
            - 'device_id': the switch's physical NIC's PnP device ID.
            - 'total_vfs': the switch's maximum number of VFs. (>= 0)
            - 'used_vfs': the switch's number of used VFs. (<= 'total_vfs')
            - 'total_iov_queue_pairs': the switch's maximum number of IOV
                queue pairs. (>= 'total_vfs')
            - 'used_iov_queue_pairs': the switch's number of used IOV queue
                pairs (<= 'total_iov_queue_pairs')
            - 'total_vmqs': the switch's maximum number of VMQs. (>= 0)
            - 'used_vmqs': the switch's number of used VMQs. (<= 'total_vmqs')
            - 'total_ipsecsa': the maximum number of IPsec SA offloads
                supported by the switch. (>= 0)
            - 'used_ipsecsa': the switch's number of IPsec SA offloads
                currently in use. (<= 'total_ipsecsa')
        """

        hw_offload_data = []

        vswitch_sds = self._conn.Msvm_VirtualEthernetSwitchSettingData()
        hw_offload_sds = self._conn.Msvm_EthernetSwitchHardwareOffloadData()
        for vswitch_sd in vswitch_sds:
            hw_offload = [
                s for s in hw_offload_sds if
                s.SystemName == vswitch_sd.VirtualSystemIdentifier][0]

            vswitch_offload_data = self._get_nic_hw_offload_info(
                vswitch_sd, hw_offload)
            if vswitch_offload_data:
                hw_offload_data.append(vswitch_offload_data)

        return hw_offload_data

    def _get_nic_hw_offload_info(self, vswitch_sd, hw_offload_sd):
        nic_name = self._netutils.get_vswitch_external_network_name(
            vswitch_sd.ElementName)
        if not nic_name:
            # NOTE(claudiub): This can happen if the vSwitch is not
            # external.
            LOG.warning("VSwitch %s is not external.", vswitch_sd.ElementName)
            return

        # check if the vSwitch is misconfigured.
        if vswitch_sd.IOVPreferred and not hw_offload_sd.IovVfCapacity:
            LOG.warning("VSwitch %s has SR-IOV enabled, but it is not "
                        "supported by the NIC or by the OS.",
                        vswitch_sd.ElementName)

        nic = self._conn_scimv2.MSFT_NetAdapter(
            InterfaceDescription=nic_name)[0]

        return {
            'vswitch_name': vswitch_sd.ElementName,
            'device_id': nic.PnPDeviceID,
            'total_vfs': hw_offload_sd.IovVfCapacity,
            'used_vfs': hw_offload_sd.IovVfUsage,
            'total_iov_queue_pairs': hw_offload_sd.IovQueuePairCapacity,
            'used_iov_queue_pairs': hw_offload_sd.IovQueuePairUsage,
            'total_vmqs': hw_offload_sd.VmqCapacity,
            'used_vmqs': hw_offload_sd.VmqUsage,
            'total_ipsecsa': hw_offload_sd.IPsecSACapacity,
            'used_ipsecsa': hw_offload_sd.IPsecSAUsage,
        }

    def get_numa_nodes(self):
        """Returns the host's list of NUMA nodes.

        :returns: list of dictionaries containing information about each
            host NUMA node. Each host has at least one NUMA node.
        """
        numa_nodes = self._conn.Msvm_NumaNode()
        nodes_info = []
        system_memory = self._conn.Msvm_Memory(['NumberOfBlocks'])
        processors = self._conn.Msvm_Processor(['DeviceID'])

        for node in numa_nodes:
            # Due to a bug in vmms, getting Msvm_Processor for the numa
            # node associators resulted in a vmms crash.
            # As an alternative to using associators we have to manually get
            # the related Msvm_Processor classes.
            # Msvm_HostedDependency is the association class between
            # Msvm_NumaNode and Msvm_Processor. We need to use this class to
            # relate the two because using associators on Msvm_Processor
            # will also result in a crash.
            numa_assoc = self._conn.Msvm_HostedDependency(
                Antecedent=node.path_())
            numa_node_assoc = [item.Dependent for item in numa_assoc]

            memory_info = self._get_numa_memory_info(numa_node_assoc,
                                                     system_memory)
            if not memory_info:
                LOG.warning("Could not find memory information for NUMA "
                            "node. Skipping node measurements.")
                continue

            cpu_info = self._get_numa_cpu_info(numa_node_assoc, processors)
            if not cpu_info:
                LOG.warning("Could not find CPU information for NUMA "
                            "node. Skipping node measurements.")
                continue

            node_info = {
                # NodeID has the format: Microsoft:PhysicalNode\<NODE_ID>
                'id': node.NodeID.split('\\')[-1],

                # memory block size is 1MB.
                'memory': memory_info.NumberOfBlocks,
                'memory_usage': node.CurrentlyConsumableMemoryBlocks,

                # DeviceID has the format: Microsoft:UUID\0\<DEV_ID>
                'cpuset': set([c.DeviceID.split('\\')[-1] for c in cpu_info]),
                # cpu_usage can be set, each CPU has a "LoadPercentage"
                'cpu_usage': 0,
            }

            nodes_info.append(node_info)

        return nodes_info

    def _get_numa_memory_info(self, numa_node_assoc, system_memory):
        memory_info = []
        paths = [x.path_().upper() for x in numa_node_assoc]
        for memory in system_memory:
            if memory.path_().upper() in paths:
                memory_info.append(memory)

        if memory_info:
            return memory_info[0]

    def _get_numa_cpu_info(self, numa_node_assoc, processors):
        cpu_info = []
        paths = [x.path_().upper() for x in numa_node_assoc]
        for proc in processors:
            if proc.path_().upper() in paths:
                cpu_info.append(proc)

        return cpu_info

    def get_remotefx_gpu_info(self):
        """Returns information about the GPUs used for RemoteFX.

        :returns: list with dictionaries containing information about each
            GPU used for RemoteFX.
        """
        gpus = []
        all_gpus = self._conn.Msvm_Physical3dGraphicsProcessor(
            EnabledForVirtualization=True)
        for gpu in all_gpus:
            gpus.append({'name': gpu.Name,
                         'driver_version': gpu.DriverVersion,
                         'total_video_ram': gpu.TotalVideoMemory,
                         'available_video_ram': gpu.AvailableVideoMemory,
                         'directx_version': gpu.DirectXVersion})
        return gpus

    def verify_host_remotefx_capability(self):
        """Validates that the host supports RemoteFX.

        :raises exceptions.HyperVRemoteFXException: if the host has no GPU
            that supports DirectX 11, or SLAT.
        """
        synth_3d_video_pool = self._conn.Msvm_Synth3dVideoPool()[0]
        if not synth_3d_video_pool.IsGpuCapable:
            raise exceptions.HyperVRemoteFXException(
                _("To enable RemoteFX on Hyper-V at least one GPU supporting "
                  "DirectX 11 is required."))
        if not synth_3d_video_pool.IsSlatCapable:
            raise exceptions.HyperVRemoteFXException(
                _("To enable RemoteFX on Hyper-V it is required that the host "
                  "GPUs support SLAT."))

    def is_host_guarded(self):
        """Checks if the host is guarded.

        :returns: False, only Windows / Hyper-V Server 2016 or newer can be
            guarded.
        """
        return False

    def supports_nested_virtualization(self):
        """Checks if the host supports nested virtualization.

        :returns: False, only Windows / Hyper-V Server 2016 or newer supports
            nested virtualization.
        """
        return False

    def get_pci_passthrough_devices(self):
        """Get host PCI devices path.

        Discrete device assignment is supported only on Windows / Hyper-V
        Server 2016 or newer.

        :returns: a list of the assignable PCI devices.
        """

        return []



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from os_win._i18n import _
from os_win import constants
from os_win import exceptions
from os_win.utils import baseutils
from os_win.utils.compute import vmutils
from os_win.utils import jobutils

LOG = logging.getLogger(__name__)


class MigrationUtils(baseutils.BaseUtilsVirt):

    def __init__(self):
        super(MigrationUtils, self).__init__()
        self._vmutils = vmutils.VMUtils()
        self._jobutils = jobutils.JobUtils()

    def _get_export_setting_data(self, vm_name):
        vm = self._vmutils._lookup_vm(vm_name)
        export_sd = self._compat_conn.Msvm_VirtualSystemExportSettingData(
            InstanceID=vm.InstanceID)
        return export_sd[0]

    def export_vm(self, vm_name, export_path,
                  copy_snapshots_config=constants.EXPORT_CONFIG_SNAPSHOTS_ALL,
                  copy_vm_storage=False, create_export_subdir=False):
        vm = self._vmutils._lookup_vm(vm_name)
        export_setting_data = self._get_export_setting_data(vm_name)

        export_setting_data.CopySnapshotConfiguration = copy_snapshots_config
        export_setting_data.CopyVmStorage = copy_vm_storage
        export_setting_data.CreateVmExportSubdirectory = create_export_subdir

        (job_path, ret_val) = self._vs_man_svc.ExportSystemDefinition(
            ComputerSystem=vm.path_(),
            ExportDirectory=export_path,
            ExportSettingData=export_setting_data.GetText_(1))
        self._jobutils.check_ret_val(ret_val, job_path)

    def import_vm_definition(self, export_config_file_path,
                             snapshot_folder_path,
                             new_uuid=False):
        (ref, job_path, ret_val) = self._vs_man_svc.ImportSystemDefinition(
            new_uuid, snapshot_folder_path, export_config_file_path)
        self._jobutils.check_ret_val(ret_val, job_path)

    def realize_vm(self, vm_name):
        planned_vm = self._get_planned_vm(vm_name, fail_if_not_found=True)

        if planned_vm:
            (job_path, ret_val) = (
                self._vs_man_svc.ValidatePlannedSystem(planned_vm.path_()))
            self._jobutils.check_ret_val(ret_val, job_path)
            (job_path, ref, ret_val) = (
                self._vs_man_svc.RealizePlannedSystem(planned_vm.path_()))
            self._jobutils.check_ret_val(ret_val, job_path)

    def _get_planned_vm(self, vm_name, conn_v2=None, fail_if_not_found=False):
        if not conn_v2:
            conn_v2 = self._conn
        planned_vm = conn_v2.Msvm_PlannedComputerSystem(ElementName=vm_name)
        if planned_vm:
            return planned_vm[0]
        elif fail_if_not_found:
            raise exceptions.HyperVException(
                _('Cannot find planned VM with name: %s') % vm_name)
        return None

    def planned_vm_exists(self, vm_name):
        """Checks if the Planned VM with the given name exists on the host."""
        return self._get_planned_vm(vm_name) is not None

    def _destroy_planned_vm(self, planned_vm):
        LOG.debug("Destroying existing planned VM: %s",
                  planned_vm.ElementName)
        (job_path,
         ret_val) = self._vs_man_svc.DestroySystem(planned_vm.path_())
        self._jobutils.check_ret_val(ret_val, job_path)

    def destroy_existing_planned_vm(self, vm_name):
        planned_vm = self._get_planned_vm(vm_name, self._compat_conn)
        if planned_vm:
            self._destroy_planned_vm(planned_vm)



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils import _wqlutils


class WqlUtilsTestCase(test_base.OsWinBaseTestCase):
    def _test_get_element_associated_class(self, fields=None):
        mock_conn = mock.MagicMock()
        _wqlutils.get_element_associated_class(
            mock_conn, mock.sentinel.class_name,
            element_instance_id=mock.sentinel.instance_id,
            fields=fields)

        expected_fields = ", ".join(fields) if fields else '*'
        expected_query = (
            "SELECT %(expected_fields)s FROM %(class_name)s "
            "WHERE InstanceID LIKE '%(instance_id)s%%'" %
            {'expected_fields': expected_fields,
             'class_name': mock.sentinel.class_name,
             'instance_id': mock.sentinel.instance_id})
        mock_conn.query.assert_called_once_with(expected_query)

    def test_get_element_associated_class(self):
        self._test_get_element_associated_class()

    def test_get_element_associated_class_specific_fields(self):
        self._test_get_element_associated_class(
            fields=['field', 'another_field'])

    def test_get_element_associated_class_invalid_element(self):
        self.assertRaises(
            exceptions.WqlException,
            _wqlutils.get_element_associated_class,
            mock.sentinel.conn,
            mock.sentinel.class_name)



# Copyright 2013 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import importutils

from os_win._i18n import _  # noqa
from os_win import exceptions
from os_win.utils import hostutils
from os_win.utils.io import namedpipe

utils = hostutils.HostUtils()

utils_map = {
    'diskutils': {
        'DiskUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.storage.diskutils.DiskUtils'}},
    'hostutils': {
        'HostUtils': {
            'min_version': 6.2,
            'max_version': 10,
            'path': 'os_win.utils.hostutils.HostUtils'},
        'HostUtils10': {
            'min_version': 10,
            'max_version': None,
            'path': 'os_win.utils.hostutils10.HostUtils10'}},
    'iscsi_initiator_utils': {
        'ISCSIInitiatorUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.storage.initiator.iscsi_utils.'
                    'ISCSIInitiatorUtils'}},
    'iscsi_target_utils': {
        'ISCSITargetUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.storage.target.iscsi_target_utils.'
                    'ISCSITargetUtils'}},
    'fc_utils': {
        'FCUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.storage.initiator.fc_utils.'
                    'FCUtils'}},
    'livemigrationutils': {
        'LiveMigrationUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.compute.livemigrationutils.'
                    'LiveMigrationUtils'}},
    'metricsutils': {
        'MetricsUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.metrics.metricsutils.MetricsUtils'}},
    'migrationutils': {
        'MigrationUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.compute.migrationutils.MigrationUtils'}},
    'networkutils': {
        'NetworkUtils': {
            'min_version': 6.2,
            'max_version': 6.3,
            'path': 'os_win.utils.network.networkutils.NetworkUtils'},
        'NetworkUtilsR2': {
            'min_version': 6.3,
            'max_version': None,
            'path': 'os_win.utils.network.networkutils.NetworkUtilsR2'}},
    'nvgreutils': {
        'NvgreUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.network.nvgreutils.NvgreUtils'}},
    'pathutils': {
        'PathUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.pathutils.PathUtils'}},
    'rdpconsoleutils': {
        'RDPConsoleUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.compute.rdpconsoleutils.RDPConsoleUtils'}},
    'smbutils': {
        'SMBUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.storage.smbutils.SMBUtils'}},
    'vhdutils': {
        'VHDUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.storage.virtdisk.vhdutils.VHDUtils'}},
    'vmutils': {
        'VMUtils': {
            'min_version': 6.2,
            'max_version': 6.3,
            'path': 'os_win.utils.compute.vmutils.VMUtils'},
        'VMUtils6_3': {
            'min_version': 6.3,
            'max_version': 10,
            'path': 'os_win.utils.compute.vmutils.VMUtils6_3'},
        'VMUtils10': {
            'min_version': 10,
            'max_version': None,
            'path': 'os_win.utils.compute.vmutils10.VMUtils10'}},
    'clusterutils': {
        'ClusterUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.compute.clusterutils.ClusterUtils'}},
    'dnsutils': {
        'DNSUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.dns.dnsutils.DNSUtils'}},
    'processutils': {
        'ProcessUtils': {
            'min_version': 6.2,
            'max_version': None,
            'path': 'os_win.utils.processutils.ProcessUtils'}}
}


def _get_class(class_type, *args, **kwargs):
    if class_type not in utils_map:
        raise exceptions.HyperVException(_('Class type %s does '
                                           'not exist') % class_type)

    windows_version = utils.get_windows_version()
    build = list(map(int, windows_version.split('.')))
    windows_version = float("%i.%i" % (build[0], build[1]))

    existing_classes = utils_map.get(class_type)
    for class_variant in existing_classes.keys():
        utils_class = existing_classes.get(class_variant)
        if (utils_class['min_version'] <= windows_version and
                (utils_class['max_version'] is None or
                 windows_version < utils_class['max_version'])):
            return importutils.import_object(utils_class['path'],
                                             *args, **kwargs)

    raise exceptions.HyperVException(_('Could not find any %(class)s class for'
                                       'this Windows version: %(win_version)s')
                                     % {'class': class_type,
                                        'win_version': windows_version})


def get_vmutils(host='.'):
    return _get_class(class_type='vmutils', host=host)


def get_vhdutils():
    return _get_class(class_type='vhdutils')


def get_metricsutils():
    return _get_class(class_type='metricsutils')


def get_networkutils():
    return _get_class(class_type='networkutils')


def get_nvgreutils():
    return _get_class(class_type='nvgreutils')


def get_hostutils():
    return _get_class(class_type='hostutils')


def get_pathutils():
    return _get_class(class_type='pathutils')


def get_iscsi_initiator_utils():
    return _get_class(class_type='iscsi_initiator_utils')


def get_livemigrationutils():
    return _get_class(class_type='livemigrationutils')


def get_smbutils():
    return _get_class(class_type='smbutils')


def get_rdpconsoleutils():
    return _get_class(class_type='rdpconsoleutils')


def get_iscsi_target_utils():
    return _get_class(class_type='iscsi_target_utils')


def get_named_pipe_handler(*args, **kwargs):
    return namedpipe.NamedPipeHandler(*args, **kwargs)


def get_fc_utils():
    return _get_class(class_type='fc_utils')


def get_diskutils():
    return _get_class(class_type='diskutils')


def get_clusterutils():
    return _get_class(class_type='clusterutils')


def get_dnsutils():
    return _get_class(class_type='dnsutils')


def get_migrationutils():
    return _get_class(class_type='migrationutils')


def get_processutils():
    return _get_class(class_type='processutils')



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import mock

from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils import hostutils10


class HostUtils10TestCase(test_base.OsWinBaseTestCase):
    """Unit tests for the Hyper-V HostUtils10 class."""

    def setUp(self):
        super(HostUtils10TestCase, self).setUp()
        self._hostutils = hostutils10.HostUtils10()
        self._hostutils._conn_hgs_attr = mock.MagicMock()
        self._hostutils._conn_attr = mock.MagicMock()
        self._hostutils._conn_cimv2 = mock.MagicMock()

    @mock.patch.object(hostutils10.HostUtils10, '_get_wmi_conn')
    def test_conn_hgs(self, mock_get_wmi_conn):
        self._hostutils._conn_hgs_attr = None
        self.assertEqual(mock_get_wmi_conn.return_value,
                         self._hostutils._conn_hgs)

        mock_get_wmi_conn.assert_called_once_with(
            self._hostutils._HGS_NAMESPACE % self._hostutils._host)

    @mock.patch.object(hostutils10.HostUtils10, '_get_wmi_conn')
    def test_conn_hgs_no_namespace(self, mock_get_wmi_conn):
        self._hostutils._conn_hgs_attr = None

        mock_get_wmi_conn.side_effect = [exceptions.OSWinException]
        self.assertRaises(exceptions.OSWinException,
                          lambda: self._hostutils._conn_hgs)
        mock_get_wmi_conn.assert_called_once_with(
            self._hostutils._HGS_NAMESPACE % self._hostutils._host)

    def _test_is_host_guarded(self, return_code=0, is_host_guarded=True):
        hgs_config = self._hostutils._conn_hgs.MSFT_HgsClientConfiguration
        hgs_config.Get.return_value = (return_code,
                                       mock.MagicMock
                                       (IsHostGuarded=is_host_guarded))
        expected_result = is_host_guarded and not return_code

        result = self._hostutils.is_host_guarded()
        self.assertEqual(expected_result, result)

    def test_is_guarded_host_config_error(self):
        self._test_is_host_guarded(return_code=mock.sentinel.return_code)

    def test_is_guarded_host(self):
        self._test_is_host_guarded()

    def test_is_not_guarded_host(self):
        self._test_is_host_guarded(is_host_guarded=False)

    def test_supports_nested_virtualization(self):
        self.assertTrue(self._hostutils.supports_nested_virtualization())

    @mock.patch.object(hostutils10.HostUtils10, '_get_pci_device_address')
    def test_get_pci_passthrough_devices(self, mock_get_pci_device_address):
        mock_pci_dev = mock.MagicMock(
            DeviceInstancePath='PCIP\\VEN_15B3&DEV_1007&SUBSYS_001815B3')
        self._hostutils._conn.Msvm_PciExpress.return_value = [mock_pci_dev] * 3
        mock_get_pci_device_address.side_effect = [
            None, mock.sentinel.address, mock.sentinel.address]

        pci_devices = self._hostutils.get_pci_passthrough_devices()

        expected_pci_dev = {
            'address': mock.sentinel.address,
            'vendor_id': '15B3',
            'product_id': '1007',
            'dev_id': mock_pci_dev.DeviceID}
        self.assertEqual([expected_pci_dev], pci_devices)
        self._hostutils._conn.Msvm_PciExpress.assert_called_once_with()
        mock_get_pci_device_address.has_calls(
            [mock.call(mock_pci_dev.DeviceInstancePath)] * 3)

    def _check_get_pci_device_address_None(self, return_code=0):
        pnp_device = mock.MagicMock()
        pnp_device.GetDeviceProperties.return_value = (
            return_code, [mock.MagicMock()])
        self._hostutils._conn_cimv2.Win32_PnPEntity.return_value = [pnp_device]

        pci_dev_address = self._hostutils._get_pci_device_address(
            mock.sentinel.pci_device_path)
        self.assertIsNone(pci_dev_address)

    def test_get_pci_device_address_error(self):
        self._check_get_pci_device_address_None(return_code=1)

    def test_get_pci_device_address_exception(self):
        self._check_get_pci_device_address_None()

    def test_get_pci_device_address(self):
        pnp_device = mock.MagicMock()
        pnp_device_properties = [
            mock.MagicMock(KeyName='DEVPKEY_Device_LocationInfo',
                           Data="bus 2, domain 4, function 0"),
            mock.MagicMock(KeyName='DEVPKEY_Device_Address',
                           Data=0)]
        pnp_device.GetDeviceProperties.return_value = (
            0, pnp_device_properties)
        self._hostutils._conn_cimv2.Win32_PnPEntity.return_value = [pnp_device]

        result = self._hostutils._get_pci_device_address(
            mock.sentinel.device_instance_path)

        pnp_props = {prop.KeyName: prop.Data for prop in pnp_device_properties}
        location_info = pnp_props['DEVPKEY_Device_LocationInfo']
        slot = pnp_props['DEVPKEY_Device_Address']
        [bus, domain, function] = re.findall(r'\b\d+\b', location_info)
        expected_result = "%04x:%02x:%02x.%1x" % (
            int(domain), int(bus), int(slot), int(function))

        self.assertEqual(expected_result, result)
        self._hostutils._conn_cimv2.Win32_PnPEntity.assert_called_once_with(
            DeviceID=mock.sentinel.device_instance_path)



#  Copyright 2013 Cloudbase Solutions Srl
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from os_win.tests.unit import test_base
from os_win.utils.compute import rdpconsoleutils


class RDPConsoleUtilsTestCase(test_base.OsWinBaseTestCase):
    _FAKE_RDP_PORT = 1000

    def setUp(self):
        self._rdpconsoleutils = rdpconsoleutils.RDPConsoleUtils()
        self._rdpconsoleutils._conn_attr = mock.MagicMock()

        super(RDPConsoleUtilsTestCase, self).setUp()

    def test_get_rdp_console_port(self):
        conn = self._rdpconsoleutils._conn
        mock_rdp_setting_data = conn.Msvm_TerminalServiceSettingData()[0]
        mock_rdp_setting_data.ListenerPort = self._FAKE_RDP_PORT

        listener_port = self._rdpconsoleutils.get_rdp_console_port()

        self.assertEqual(self._FAKE_RDP_PORT, listener_port)



# Copyright 2017 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import textwrap

import mock
import pep8

from os_win._hacking import checks
from os_win.tests.unit import test_base


class HackingTestCase(test_base.OsWinBaseTestCase):
    """This class tests the hacking checks in os_win.hacking.checks.

    This is accomplished by passing strings to the check methods like the
    pep8/flake8 parser would. The parser loops over each line in the file and
    then passes the parameters to the check method. The parameter names in the
    check method dictate what type of object is passed to the check method.

    The parameter types are:
        logical_line: A processed line with the following modifications:
            - Multi-line statements converted to a single line.
            - Stripped left and right.
            - Contents of strings replaced with "xxx" of same length.
            - Comments removed.
        physical_line: Raw line of text from the input file.
        lines: a list of the raw lines from the input file
        tokens: the tokens that contribute to this logical line
        line_number: line number in the input file
        total_lines: number of lines in the input file
        blank_lines: blank lines before this one
        indent_char: indentation character in this file (" " or "\t")
        indent_level: indentation (with tabs expanded to multiples of 8)
        previous_indent_level: indentation on previous line
        previous_logical: previous logical line
        filename: Path of the file being run through pep8

    When running a test on a check method the return will be False/None if
    there is no violation in the sample input. If there is an error a tuple is
    returned with a position in the line, and a message. So to check the result
    just assertTrue if the check is expected to fail and assertFalse if it
    should pass.
    """

    def _run_check(self, code, checker, filename=None):
        # We are patching pep8 so that only the check under test is actually
        # installed.
        mock_checks = {'physical_line': {}, 'logical_line': {}, 'tree': {}}
        with mock.patch('pep8._checks', mock_checks):
            pep8.register_check(checker)

            lines = textwrap.dedent(code).strip().splitlines(True)

            checker = pep8.Checker(filename=filename, lines=lines)
            # NOTE(sdague): the standard reporter has printing to stdout
            # as a normal part of check_all, which bleeds through to the
            # test output stream in an unhelpful way. This blocks that
            # printing.
            with mock.patch('pep8.StandardReport.get_file_results'):
                checker.check_all()
            checker.report._deferred_print.sort()
            return checker.report._deferred_print

    def _assert_has_errors(self, code, checker, expected_errors=None,
                           filename=None):
        actual_errors = [e[:3] for e in
                         self._run_check(code, checker, filename)]
        self.assertEqual(expected_errors or [], actual_errors)

    def _assert_has_no_errors(self, code, checker, filename=None):
        self._assert_has_errors(code, checker, filename=filename)

    def test_ctypes_libs_not_used_directly(self):
        checker = checks.assert_ctypes_libs_not_used_directly
        errors = [(1, 0, 'O301')]

        code = "ctypes.cdll.hbaapi"
        self._assert_has_errors(code, checker, expected_errors=errors)

        code = "ctypes.windll.hbaapi.fake_func(fake_arg)"
        self._assert_has_errors(code, checker, expected_errors=errors)

        code = "fake_var = ctypes.oledll.hbaapi.fake_func(fake_arg)"
        self._assert_has_errors(code, checker, expected_errors=errors)

        code = "foo(ctypes.pydll.hbaapi.fake_func(fake_arg))"
        self._assert_has_errors(code, checker, expected_errors=errors)

        code = "ctypes.cdll.LoadLibrary(fake_lib)"
        self._assert_has_errors(code, checker, expected_errors=errors)

        code = "ctypes.WinDLL('fake_lib_path')"
        self._assert_has_errors(code, checker, expected_errors=errors)

        code = "ctypes.cdll.hbaapi"
        filename = os.path.join("os_win", "utils", "winapi",
                                "libs", "hbaapi.py")
        self._assert_has_no_errors(code, checker, filename=filename)

    def test_ctypes_foreign_func_argtypes_defined(self):
        checker = checks.assert_ctypes_foreign_func_argtypes_defined
        errors = [(1, 0, 'O302')]

        code = "kernel32.FakeFunc(fake_arg)"
        self._assert_has_errors(code, checker, errors)

        code = "fake_func(kernel32.FakeFunc(fake_arg))"
        self._assert_has_errors(code, checker, errors)

        code = "kernel32.WaitNamedPipeW(x, y)"
        self._assert_has_no_errors(code, checker)

        code = "_fake_kernel32.WaitNamedPipeW(x, y)"
        self._assert_has_no_errors(code, checker)

    def test_no_log_translations(self):
        for log in checks._all_log_levels:
            bad = 'LOG.%s(_("Bad"))' % log
            self.assertEqual(1, len(list(checks.no_translate_logs(bad))))
            # Catch abuses when used with a variable and not a literal
            bad = 'LOG.%s(_(msg))' % log
            self.assertEqual(1, len(list(checks.no_translate_logs(bad))))



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from oslo_log import log as logging
import six

from os_win._i18n import _
from os_win import _utils
from os_win import constants
from os_win import exceptions
from os_win.utils import _wqlutils
from os_win.utils.compute import vmutils
from oslo_utils import units

LOG = logging.getLogger(__name__)


class VMUtils10(vmutils.VMUtils6_3):

    _UEFI_CERTIFICATE_AUTH = 'MicrosoftUEFICertificateAuthority'
    _SERIAL_PORT_SETTING_DATA_CLASS = "Msvm_SerialPortSettingData"
    _SECURITY_SETTING_DATA = 'Msvm_SecuritySettingData'
    _PCI_EXPRESS_SETTING_DATA = 'Msvm_PciExpressSettingData'
    _MSPS_NAMESPACE = '//%s/root/msps'

    _remote_fx_res_map = {
        constants.REMOTEFX_MAX_RES_1024x768: 0,
        constants.REMOTEFX_MAX_RES_1280x1024: 1,
        constants.REMOTEFX_MAX_RES_1600x1200: 2,
        constants.REMOTEFX_MAX_RES_1920x1200: 3,
        constants.REMOTEFX_MAX_RES_2560x1600: 4,
        constants.REMOTEFX_MAX_RES_3840x2160: 5
    }

    _remotefx_max_monitors_map = {
        # defines the maximum number of monitors for a given
        # resolution
        constants.REMOTEFX_MAX_RES_1024x768: 8,
        constants.REMOTEFX_MAX_RES_1280x1024: 8,
        constants.REMOTEFX_MAX_RES_1600x1200: 4,
        constants.REMOTEFX_MAX_RES_1920x1200: 4,
        constants.REMOTEFX_MAX_RES_2560x1600: 2,
        constants.REMOTEFX_MAX_RES_3840x2160: 1
    }

    _remotefx_vram_vals = [64 * units.Mi, 128 * units.Mi, 256 * units.Mi,
                           512 * units.Mi, 1024 * units.Mi]

    def __init__(self, host='.'):
        super(VMUtils10, self).__init__(host)
        self._conn_msps_attr = None
        self._sec_svc_attr = None

    @property
    def _conn_msps(self):
        if not self._conn_msps_attr:
            try:
                namespace = self._MSPS_NAMESPACE % self._host
                self._conn_msps_attr = self._get_wmi_conn(namespace)
            except Exception:
                raise exceptions.OSWinException(
                    _("Namespace %(namespace)s not found. Make sure "
                      "FabricShieldedTools feature is installed.") %
                    {'namespace': namespace})

        return self._conn_msps_attr

    @property
    def _sec_svc(self):
        if not self._sec_svc_attr:
            self._sec_svc_attr = self._conn.Msvm_SecurityService()[0]
        return self._sec_svc_attr

    def set_nested_virtualization(self, vm_name, state):
        """Enables nested virtualization for the given VM.

        :param vm_name: the name of the VM.
        :param state: boolean, if True, nested virtualization will be enabled,
            disabled otherwise.
        """
        vmsettings = self._lookup_vm_check(vm_name)
        procsettings = _wqlutils.get_element_associated_class(
            self._conn, self._PROCESSOR_SETTING_DATA_CLASS,
            element_instance_id=vmsettings.InstanceID)[0]

        procsettings.ExposeVirtualizationExtensions = state
        self._jobutils.modify_virt_resource(procsettings)

    def vm_gen_supports_remotefx(self, vm_gen):
        """RemoteFX is supported on both generation 1 and 2 virtual

        machines for Windows 10 / Windows Server 2016.

        :returns: True
        """

        return True

    def _validate_remotefx_params(self, monitor_count, max_resolution,
                                  vram_bytes=None):
        super(VMUtils10, self)._validate_remotefx_params(monitor_count,
                                                         max_resolution)
        if vram_bytes and vram_bytes not in self._remotefx_vram_vals:
            raise exceptions.HyperVRemoteFXException(
                _("Unsuported RemoteFX VRAM value: %(requested_value)s."
                  "The supported VRAM values are: %(supported_values)s") %
                {'requested_value': vram_bytes,
                 'supported_values': self._remotefx_vram_vals})

    def _set_remotefx_vram(self, remotefx_disp_ctrl_res, vram_bytes):
        if vram_bytes:
            remotefx_disp_ctrl_res.VRAMSizeBytes = six.text_type(vram_bytes)

    def _vm_has_s3_controller(self, vm_name):
        return self.get_vm_generation(vm_name) == constants.VM_GEN_1

    def _set_secure_boot(self, vs_data, msft_ca_required):
        vs_data.SecureBootEnabled = True
        if msft_ca_required:
            uefi_data = self._conn.Msvm_VirtualSystemSettingData(
                ElementName=self._UEFI_CERTIFICATE_AUTH)[0]
            vs_data.SecureBootTemplateId = uefi_data.SecureBootTemplateId

    def populate_fsk(self, fsk_filepath, fsk_pairs):
        """Writes in the fsk file all the substitution strings and their

        values which will populate the unattended file used when
        creating the pdk.
        """

        fabric_data_pairs = []
        for fsk_key, fsk_value in fsk_pairs.items():
            fabricdata = self._conn_msps.Msps_FabricData.new()
            fabricdata.key = fsk_key
            fabricdata.Value = fsk_value
            fabric_data_pairs.append(fabricdata)

        fsk = self._conn_msps.Msps_FSK.new()
        fsk.FabricDataPairs = fabric_data_pairs
        msps_pfp = self._conn_msps.Msps_ProvisioningFileProcessor

        msps_pfp.SerializeToFile(fsk_filepath, fsk)

    def add_vtpm(self, vm_name, pdk_filepath, shielded):
        """Adds a vtpm and enables it with encryption or shielded option."""

        vm = self._lookup_vm_check(vm_name)

        msps_pfp = self._conn_msps.Msps_ProvisioningFileProcessor
        provisioning_file = msps_pfp.PopulateFromFile(pdk_filepath)[0]
        # key_protector: array of bytes
        key_protector = provisioning_file.KeyProtector
        # policy_data: array of bytes
        policy_data = provisioning_file.PolicyData

        security_profile = _wqlutils.get_element_associated_class(
            self._conn, self._SECURITY_SETTING_DATA,
            element_uuid=vm.ConfigurationID)[0]

        security_profile.EncryptStateAndVmMigrationTraffic = True
        security_profile.TpmEnabled = True
        security_profile.ShieldingRequested = shielded

        sec_profile_serialized = security_profile.GetText_(1)
        (job_path, ret_val) = self._sec_svc.SetKeyProtector(
            key_protector, sec_profile_serialized)
        self._jobutils.check_ret_val(ret_val, job_path)

        (job_path, ret_val) = self._sec_svc.SetSecurityPolicy(
            policy_data, sec_profile_serialized)
        self._jobutils.check_ret_val(ret_val, job_path)

        (job_path, ret_val) = self._sec_svc.ModifySecuritySettings(
            sec_profile_serialized)
        self._jobutils.check_ret_val(ret_val, job_path)

    def provision_vm(self, vm_name, fsk_filepath, pdk_filepath):
        vm = self._lookup_vm_check(vm_name)
        provisioning_service = self._conn_msps.Msps_ProvisioningService

        (job_path, ret_val) = provisioning_service.ProvisionMachine(
            fsk_filepath, vm.ConfigurationID, pdk_filepath)
        self._jobutils.check_ret_val(ret_val, job_path)

    def is_secure_vm(self, instance_name):
        inst_id = self.get_vm_id(instance_name)
        security_profile = _wqlutils.get_element_associated_class(
            self._conn, self._SECURITY_SETTING_DATA,
            element_uuid=inst_id)
        if security_profile:
            return security_profile[0].EncryptStateAndVmMigrationTraffic
        return False

    def add_pci_device(self, vm_name, vendor_id, product_id):
        """Adds the given PCI device to the given VM.

        :param vm_name: the name of the VM to which the PCI device will be
            attached to.
        :param vendor_id: the PCI device's vendor ID.
        :param product_id: the PCI device's product ID.
        :raises exceptions.PciDeviceNotFound: if there is no PCI device
            identifiable by the given vendor_id and product_id, or it was
            already assigned.
        """
        vmsettings = self._lookup_vm_check(vm_name)
        pci_setting_data = self._get_new_setting_data(
            self._PCI_EXPRESS_SETTING_DATA)
        pci_device = self._get_assignable_pci_device(vendor_id, product_id)
        pci_setting_data.HostResource = [pci_device.path_()]

        self._jobutils.add_virt_resource(pci_setting_data, vmsettings)

    def _get_assignable_pci_device(self, vendor_id, product_id):
        pci_devices = self._conn.Msvm_PciExpress()

        pattern = re.compile(
            "^(.*)VEN_%(vendor_id)s&DEV_%(product_id)s&(.*)$" % {
                'vendor_id': vendor_id, 'product_id': product_id})
        for dev in pci_devices:
            if pattern.match(dev.DeviceID):
                # NOTE(claudiub): if the given PCI device is already assigned,
                # the pci_devices list will contain PCI device with the same
                # LocationPath.
                pci_devices_found = [d for d in pci_devices if
                                     d.LocationPath == dev.LocationPath]

                LOG.debug('PCI devices found: %s',
                          [d.DeviceID for d in pci_devices_found])

                # device is not in use by other VM
                if len(pci_devices_found) == 1:
                    return pci_devices_found[0]

        raise exceptions.PciDeviceNotFound(vendor_id=vendor_id,
                                           product_id=product_id)

    def remove_pci_device(self, vm_name, vendor_id, product_id):
        """Removes the given PCI device from the given VM.

        :param vm_name: the name of the VM from which the PCI device will be
            attached from.
        :param vendor_id: the PCI device's vendor ID.
        :param product_id: the PCI device's product ID.
        """
        vmsettings = self._lookup_vm_check(vm_name)

        pattern = re.compile(
            "^(.*)VEN_%(vendor_id)s&DEV_%(product_id)s&(.*)$" % {
                'vendor_id': vendor_id, 'product_id': product_id})

        pci_sds = _wqlutils.get_element_associated_class(
            self._conn, self._PCI_EXPRESS_SETTING_DATA,
            vmsettings.InstanceID)
        pci_sds = [sd for sd in pci_sds if pattern.match(sd.HostResource[0])]

        if pci_sds:
            self._jobutils.remove_virt_resource(pci_sds[0])
        else:
            LOG.debug("PCI device with vendor ID %(vendor_id)s and "
                      "%(product_id)s is not attached to %(vm_name)s",
                      {'vendor_id': vendor_id, 'product_id': product_id,
                       'vm_name': vm_name})

    def remove_all_pci_devices(self, vm_name):
        """Removes all the PCI devices from the given VM.

        :param vm_name: the name of the VM from which all the PCI devices will
            be detached from.
        """
        vmsettings = self._lookup_vm_check(vm_name)

        pci_sds = _wqlutils.get_element_associated_class(
            self._conn, self._PCI_EXPRESS_SETTING_DATA,
            vmsettings.InstanceID)

        if pci_sds:
            self._jobutils.remove_multiple_virt_resources(pci_sds)

    @_utils.required_vm_version(min_version=constants.VM_VERSION_6_2)
    def _set_vm_snapshot_type(self, vmsettings, snapshot_type):
        # We expect the caller to actually push the vmsettings update.
        vmsettings.UserSnapshotType = snapshot_type



#  Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes
import os
import shutil

import mock

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils import pathutils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi.libs import advapi32 as advapi32_def


class PathUtilsTestCase(test_base.OsWinBaseTestCase):
    """Unit tests for the Hyper-V PathUtils class."""

    _autospec_classes = [
        pathutils.win32utils.Win32Utils,
        pathutils._acl_utils.ACLUtils,
    ]

    def setUp(self):
        super(PathUtilsTestCase, self).setUp()
        self._setup_lib_mocks()

        self._pathutils = pathutils.PathUtils()
        self._mock_run = self._pathutils._win32_utils.run_and_check_output
        self._acl_utils = self._pathutils._acl_utils

    def _setup_lib_mocks(self):
        self._ctypes = mock.Mock()
        self._wintypes = mock.Mock()

        self._wintypes.BOOL = lambda x: (x, 'BOOL')
        self._ctypes.c_wchar_p = lambda x: (x, "c_wchar_p")
        self._ctypes.pointer = lambda x: (x, 'pointer')

        self._ctypes_patcher = mock.patch.object(
            pathutils, 'ctypes', new=self._ctypes)
        self._ctypes_patcher.start()

        mock.patch.multiple(pathutils,
                            wintypes=self._wintypes,
                            kernel32=mock.DEFAULT,
                            create=True).start()

    @mock.patch.object(pathutils.PathUtils, 'copy')
    @mock.patch.object(os.path, 'isfile')
    @mock.patch.object(os, 'listdir')
    @mock.patch.object(pathutils.PathUtils, 'check_create_dir')
    def test_copy_folder_files(self, mock_check_create_dir, mock_listdir,
                               mock_isfile, mock_copy):
        src_dir = 'src'
        dest_dir = 'dest'
        fname = 'tmp_file.txt'
        subdir = 'tmp_folder'
        src_fname = os.path.join(src_dir, fname)
        dest_fname = os.path.join(dest_dir, fname)

        # making sure src_subdir is not copied.
        mock_listdir.return_value = [fname, subdir]
        mock_isfile.side_effect = [True, False]

        self._pathutils.copy_folder_files(src_dir, dest_dir)

        mock_check_create_dir.assert_called_once_with(dest_dir)
        mock_copy.assert_called_once_with(src_fname, dest_fname)

    @mock.patch.object(pathutils.PathUtils, 'rename')
    @mock.patch.object(os.path, 'isfile')
    @mock.patch.object(os, 'listdir')
    def test_move_folder_files(self, mock_listdir, mock_isfile, mock_rename):
        src_dir = 'src'
        dest_dir = 'dest'
        fname = 'tmp_file.txt'
        subdir = 'tmp_folder'
        src_fname = os.path.join(src_dir, fname)
        dest_fname = os.path.join(dest_dir, fname)

        # making sure src_subdir is not moved.
        mock_listdir.return_value = [fname, subdir]
        mock_isfile.side_effect = [True, False]

        self._pathutils.move_folder_files(src_dir, dest_dir)
        mock_rename.assert_called_once_with(src_fname, dest_fname)

    @mock.patch('time.sleep')
    @mock.patch.object(pathutils.shutil, 'rmtree')
    def test_rmtree(self, mock_rmtree, mock_sleep):
        exc = exceptions.WindowsError()
        exc.winerror = w_const.ERROR_DIR_IS_NOT_EMPTY
        mock_rmtree.side_effect = [exc] * 5 + [None]

        self._pathutils.rmtree(mock.sentinel.FAKE_PATH)

        mock_rmtree.assert_has_calls([mock.call(mock.sentinel.FAKE_PATH)] * 6)

    @mock.patch('time.sleep')
    @mock.patch.object(pathutils.shutil, 'rmtree')
    def _check_rmtree(self, mock_rmtree, mock_sleep, side_effect):
        mock_rmtree.side_effect = side_effect
        self.assertRaises(exceptions.OSWinException, self._pathutils.rmtree,
                          mock.sentinel.FAKE_PATH)

    def test_rmtree_unexpected(self):
        self._check_rmtree(side_effect=exceptions.WindowsError)

    def test_rmtree_exceeded(self):
        exc = exceptions.WindowsError()
        exc.winerror = w_const.ERROR_DIR_IS_NOT_EMPTY
        self._check_rmtree(side_effect=[exc] * 6)

    @mock.patch.object(pathutils.PathUtils, 'makedirs')
    @mock.patch.object(pathutils.PathUtils, 'exists')
    def test_check_create_dir(self, mock_exists, mock_makedirs):
        fake_dir = 'dir'
        mock_exists.return_value = False
        self._pathutils.check_create_dir(fake_dir)

        mock_exists.assert_called_once_with(fake_dir)
        mock_makedirs.assert_called_once_with(fake_dir)

    @mock.patch.object(pathutils.PathUtils, 'rmtree')
    @mock.patch.object(pathutils.PathUtils, 'exists')
    def test_check_remove_dir(self, mock_exists, mock_rmtree):
        fake_dir = 'dir'
        self._pathutils.check_remove_dir(fake_dir)

        mock_exists.assert_called_once_with(fake_dir)
        mock_rmtree.assert_called_once_with(fake_dir)

    @mock.patch('os.path.isdir')
    @mock.patch('os.path.islink')
    def _test_check_symlink(self, mock_is_symlink, mock_is_dir,
                            is_symlink=True, python_version=(2, 7),
                            is_dir=True):
        fake_path = r'c:\\fake_path'
        if is_symlink:
            f_attr = 0x400
        else:
            f_attr = 0x80

        mock_is_dir.return_value = is_dir
        mock_is_symlink.return_value = is_symlink
        self._mock_run.return_value = f_attr

        with mock.patch('sys.version_info', python_version):
            ret_value = self._pathutils.is_symlink(fake_path)

        if python_version >= (3, 2):
            mock_is_symlink.assert_called_once_with(fake_path)
        else:
            self._mock_run.assert_called_once_with(
                pathutils.kernel32.GetFileAttributesW,
                fake_path,
                error_ret_vals=[w_const.INVALID_FILE_ATTRIBUTES],
                kernel32_lib_func=True)

        self.assertEqual(is_symlink, ret_value)

    def test_is_symlink(self):
        self._test_check_symlink()

    def test_is_not_symlink(self):
        self._test_check_symlink(is_symlink=False)

    def test_is_symlink_python_gt_3_2(self):
        self._test_check_symlink(python_version=(3, 3))

    def test_create_sym_link(self):
        tg_is_dir = False
        self._pathutils.create_sym_link(mock.sentinel.path,
                                        mock.sentinel.target,
                                        target_is_dir=tg_is_dir)

        self._mock_run.assert_called_once_with(
            pathutils.kernel32.CreateSymbolicLinkW,
            mock.sentinel.path,
            mock.sentinel.target,
            tg_is_dir,
            kernel32_lib_func=True)

    @mock.patch('os.path.isdir')
    def _test_copy(self, mock_isdir, dest_isdir=False):
        mock_isdir.return_value = dest_isdir
        fail_if_exists = False

        fake_src = r'fake_src_fname'
        fake_dest = r'fake_dest'

        expected_dest = (os.path.join(fake_dest, fake_src)
                         if dest_isdir else fake_dest)

        self._pathutils.copy(fake_src, fake_dest,
                             fail_if_exists=fail_if_exists)

        self._mock_run.assert_called_once_with(
            pathutils.kernel32.CopyFileW,
            self._ctypes.c_wchar_p(fake_src),
            self._ctypes.c_wchar_p(expected_dest),
            self._wintypes.BOOL(fail_if_exists),
            kernel32_lib_func=True)

    def test_copy_dest_is_fpath(self):
        self._test_copy()

    def test_copy_dest_is_dir(self):
        self._test_copy(dest_isdir=True)

    @mock.patch('os.path.isdir')
    def test_copy_exc(self, mock_isdir):
        mock_isdir.return_value = False
        self._mock_run.side_effect = exceptions.Win32Exception(
            func_name='mock_copy',
            error_code='fake_error_code',
            error_message='fake_error_msg')
        self.assertRaises(IOError,
                          self._pathutils.copy,
                          mock.sentinel.src,
                          mock.sentinel.dest)

    @mock.patch('os.close')
    @mock.patch('tempfile.mkstemp')
    def test_create_temporary_file(self, mock_mkstemp, mock_close):
        fd = mock.sentinel.file_descriptor
        path = mock.sentinel.absolute_pathname
        mock_mkstemp.return_value = (fd, path)

        output = self._pathutils.create_temporary_file(
            suffix=mock.sentinel.suffix)

        self.assertEqual(path, output)
        mock_close.assert_called_once_with(fd)
        mock_mkstemp.assert_called_once_with(suffix=mock.sentinel.suffix)

    @mock.patch('oslo_utils.fileutils.delete_if_exists')
    def test_temporary_file(self, mock_delete):
        self._pathutils.create_temporary_file = mock.MagicMock()
        self._pathutils.create_temporary_file.return_value = (
            mock.sentinel.temporary_file)
        with self._pathutils.temporary_file() as tmp_file:
            self.assertEqual(mock.sentinel.temporary_file, tmp_file)
            self.assertFalse(mock_delete.called)
        mock_delete.assert_called_once_with(mock.sentinel.temporary_file)

    @mock.patch.object(shutil, 'copytree')
    @mock.patch('os.path.abspath')
    def test_copy_dir(self, mock_abspath, mock_copytree):
        mock_abspath.side_effect = [mock.sentinel.src, mock.sentinel.dest]
        self._pathutils.copy_dir(mock.sentinel.src, mock.sentinel.dest)

        mock_abspath.has_calls(
            [mock.call(mock.sentinel.src), mock.call(mock.sentinel.dest)])
        mock_copytree.assert_called_once_with(mock.sentinel.src,
                                              mock.sentinel.dest)

    def test_add_acl_rule(self):
        # We raise an expected exception in order to
        # easily verify the resource cleanup.
        raised_exc = exceptions.OSWinException
        self._ctypes_patcher.stop()

        fake_trustee = 'FAKEDOMAIN\\FakeUser'
        mock_sec_info = dict(pp_sec_desc=mock.Mock(),
                             pp_dacl=mock.Mock())
        self._acl_utils.get_named_security_info.return_value = mock_sec_info
        self._acl_utils.set_named_security_info.side_effect = raised_exc
        pp_new_dacl = self._acl_utils.set_entries_in_acl.return_value

        self.assertRaises(raised_exc,
                          self._pathutils.add_acl_rule,
                          path=mock.sentinel.path,
                          trustee_name=fake_trustee,
                          access_rights=constants.ACE_GENERIC_READ,
                          access_mode=constants.ACE_GRANT_ACCESS,
                          inheritance_flags=constants.ACE_OBJECT_INHERIT)

        self._acl_utils.get_named_security_info.assert_called_once_with(
            obj_name=mock.sentinel.path,
            obj_type=w_const.SE_FILE_OBJECT,
            security_info_flags=w_const.DACL_SECURITY_INFORMATION)
        self._acl_utils.set_entries_in_acl.assert_called_once_with(
            entry_count=1,
            p_explicit_entry_list=mock.ANY,
            p_old_acl=mock_sec_info['pp_dacl'].contents)
        self._acl_utils.set_named_security_info.assert_called_once_with(
            obj_name=mock.sentinel.path,
            obj_type=w_const.SE_FILE_OBJECT,
            security_info_flags=w_const.DACL_SECURITY_INFORMATION,
            p_dacl=pp_new_dacl.contents)

        p_access = self._acl_utils.set_entries_in_acl.call_args_list[0][1][
            'p_explicit_entry_list']
        access = ctypes.cast(
            p_access,
            ctypes.POINTER(advapi32_def.EXPLICIT_ACCESS)).contents

        self.assertEqual(constants.ACE_GENERIC_READ,
                         access.grfAccessPermissions)
        self.assertEqual(constants.ACE_GRANT_ACCESS,
                         access.grfAccessMode)
        self.assertEqual(constants.ACE_OBJECT_INHERIT,
                         access.grfInheritance)
        self.assertEqual(w_const.TRUSTEE_IS_NAME,
                         access.Trustee.TrusteeForm)
        self.assertEqual(fake_trustee,
                         access.Trustee.pstrName)

        self._pathutils._win32_utils.local_free.assert_has_calls(
            [mock.call(pointer)
             for pointer in [mock_sec_info['pp_sec_desc'].contents,
                             pp_new_dacl.contents]])

    def test_copy_acls(self):
        raised_exc = exceptions.OSWinException

        mock_sec_info = dict(pp_sec_desc=mock.Mock(),
                             pp_dacl=mock.Mock())
        self._acl_utils.get_named_security_info.return_value = mock_sec_info
        self._acl_utils.set_named_security_info.side_effect = raised_exc

        self.assertRaises(raised_exc,
                          self._pathutils.copy_acls,
                          mock.sentinel.src,
                          mock.sentinel.dest)

        self._acl_utils.get_named_security_info.assert_called_once_with(
            obj_name=mock.sentinel.src,
            obj_type=w_const.SE_FILE_OBJECT,
            security_info_flags=w_const.DACL_SECURITY_INFORMATION)
        self._acl_utils.set_named_security_info.assert_called_once_with(
            obj_name=mock.sentinel.dest,
            obj_type=w_const.SE_FILE_OBJECT,
            security_info_flags=w_const.DACL_SECURITY_INFORMATION,
            p_dacl=mock_sec_info['pp_dacl'].contents)

        self._pathutils._win32_utils.local_free.assert_called_once_with(
            mock_sec_info['pp_sec_desc'].contents)
